from typing import Optional, Union

from bokeh.model import Model
from bokeh.models import Div, Select

from neuralib.dashboard import BokehServer, View
from .view_figure import AnimalFigureView

__all__ = ['AnimalView']


class AnimalView(View):
    select: Select
    content: Div

    def __init__(self, animal: str = None):
        self._animal: Union[None, str, BaseException] = animal

    @property
    def title(self) -> str:
        animal = self.current_animal or 'Unknown'
        return f'Animal {animal}'

    @property
    def current_animal(self) -> Optional[str]:
        if self._animal is None:
            try:
                self._animal = self.get_arg('animal')[0]
            except (KeyError, IOError) as e:
                self._animal = e

        if isinstance(self._animal, str):
            return self._animal
        else:
            return None

    def setup(self) -> Model:
        self.select = Select(
            title='Figure Type',
            value='',
            options=['', *AnimalFigureView.FIGURE_TYPE]
        )
        self.select.on_change('value', self.on_figure_type)

        self.content = Div(text="", css_classes=['yw-animal-content'])

        from bokeh.layouts import column
        return column(
            self.select,
            self.content,
            Div(text="""
<style>
div.yw-animal-content div {
    position: fixed;
    width: 90%;
    height: 100%;
}
div.yw-animal-content div iframe.bk {
    display: block;
    width: 100%;
    height: 100%;
}
</style>
""")
        )

    def on_figure_type(self, attr: str, old: str, value: str):
        if len(value) == 0:
            self.content.text = ''
            return

        animal = self.current_animal
        if animal is None:
            return

        self.content.text = f""" <iframe src="/pool/figure?animal={animal}&figure={value}" class="bk" ></iframe> """


def main():
    import argparse
    ap = argparse.ArgumentParser()
    ap.add_argument('--animal', metavar='NAME', required=True)
    opt = ap.parse_args()
    BokehServer().start(AnimalView(opt.animal))


if __name__ == '__main__':
    main()



"""
SQL help functions.
"""
from __future__ import annotations

from collections.abc import Collection
from typing import overload, Any, TypeVar, TYPE_CHECKING

from . import expr

if TYPE_CHECKING:
    from .stat import SqlStat, SqlSelectStat

__all__ = [
    'TRUE', 'FALSE', 'NULL', 'ROWID',
    'literal', 'wrap', 'alias', 'cast', 'case', 'exists', 'asc', 'desc', 'nulls_first', 'nulls_last', 'concat',
    'and_', 'or_',
    'like', 'not_like', 'glob', 'contains', 'not_contains', 'between', 'not_between', 'is_null', 'is_not_null',
    'excluded', 'with_common_table', 'fields'
]

T = TypeVar('T')

TRUE = expr.SqlLiteral.SQL_TRUE
FALSE = expr.SqlLiteral.SQL_FALSE
NULL = expr.SqlLiteral.SQL_NULL
ROWID = expr.SqlLiteral('rowid')


def literal(x: str) -> expr.SqlExpr:
    return expr.SqlLiteral(x)


def wrap(x) -> expr.SqlExpr:
    """wrap *x* as an SQL expression."""
    return expr.wrap(x)


@overload
def alias(x: str, name: str) -> Any:
    pass


@overload
def alias(x: type[T], name: str) -> type[T]:
    pass


@overload
def alias(x: SqlStat[T], name: str) -> type[T]:
    pass


def alias(x, name: str):
    """
    ..  code-block::SQL

        :x AS :name

    """
    if isinstance(x, type):
        return expr.SqlAlias(x, name)
    return expr.SqlAlias(expr.wrap(x), name)


def cast(t: type[T], x) -> T:
    """
    ..  code-block::SQL

        CAST(:x AS :t)

    """
    return expr.SqlCastOper(t.__name__, expr.wrap(x))


def case(x=None) -> expr.SqlCaseExpr:
    """
    https://www.sqlite.org/lang_expr.html#the_case_expression

    .. code-block::SQL

        CASE :x
            ...

    :param x:
    :return:
    """
    return expr.SqlCaseExpr(None if x is None else expr.wrap(x))


@overload
def exists(x: SqlStat) -> expr.SqlExpr:
    pass


@overload
def exists(x: type, *where: bool | expr.SqlExpr) -> expr.SqlExpr:
    pass


def exists(x, *where: bool | expr.SqlExpr) -> expr.SqlExpr:
    """

    https://www.sqlite.org/lang_expr.html#the_exists_operator

    >>> exists(A, A.a == 1) # equivalent below
    >>> exists(select_from(1, from_table=A).where(A.a == 1))

    :param x:
    :param where:
    :return:
    """
    if isinstance(x, type):
        from .stat_start import select_from
        x = select_from(1, from_table=x).where(*where)
    return expr.SqlExistsOper('EXISTS', x)

def asc(x) -> expr.SqlExpr:
    """
    ascending ordering Used by **ORDER BY**.

    ..  code-block::SQL

        :x ASC


    """
    return expr.SqlOrderOper('ASC', expr.wrap(x))


def desc(x) -> expr.SqlExpr:
    """
    descending ordering used by **ORDER BY**.

    ..  code-block::SQL

        :x DESC

    """
    return expr.SqlOrderOper('DESC', expr.wrap(x))


def nulls_first(x) -> expr.SqlExpr:
    """
    order null first used by **ORDER BY**.

    ..  code-block::SQL

        :x NULLS FIRST

    """
    return expr.SqlOrderOper('NULLS FIRST', expr.wrap(x))


def nulls_last(x) -> expr.SqlExpr:
    """
    order null last used by **ORDER BY**.

    ..  code-block::SQL

        :x NULLS LAST

    """
    return expr.SqlOrderOper('NULLS LAST', expr.wrap(x))


def concat(*x) -> expr.SqlExpr:
    """
    concatenate strings.

    ..  code-block::SQL

        :x[0] || :x[1] || ...

    """
    return expr.SqlConcatOper(expr.wrap_seq(*x))


def and_(*other) -> expr.SqlExpr:
    """
    "AND" SQL expressions.

    ..  code-block::SQL

        other[0] AND other[1] AND ...

    """
    if len(other) == 0:
        raise RuntimeError()
    elif len(other) == 1:
        return expr.wrap(other[0])
    return expr.SqlVarArgOper('AND', expr.wrap_seq(*other))


def or_(*other) -> expr.SqlExpr:
    """
    "OR" SQL expressions.

    ..  code-block::SQL

        other[0] OR other[1] OR ...

    """
    if len(other) == 0:
        raise RuntimeError()
    elif len(other) == 1:
        return expr.wrap(other[0])
    return expr.SqlVarArgOper('OR', expr.wrap_seq(*other))


def like(x, s) -> expr.SqlCompareOper:
    """
    ..  code-block::SQL

        :x LIKE :s


    :param x:
    :param s:
    :return:
    """
    return expr.wrap(x).like(s)


def not_like(x, s) -> expr.SqlCompareOper:
    """
    ..  code-block::SQL

        :x NOT LIKE :s


    :param x:
    :param s:
    :return:
    """
    return expr.wrap(x).not_like(s)


def glob(x, s) -> expr.SqlCompareOper:
    """
    ..  code-block::SQL

        :x GLOB :s


    :param x:
    :param s:
    :return:
    """
    return expr.wrap(x).glob(s)


def contains(x, coll) -> expr.SqlCompareOper:
    """
    ..  code-block::SQL

        :x IN :COLL


    :param x:
    :param coll: a sequence or a select statement.
    :return:
    """
    from .stat import SqlStat
    if not isinstance(coll, (tuple, list, SqlStat)):
        coll = [coll]
    return expr.wrap(x).contains(coll)


def not_contains(x, coll) -> expr.SqlCompareOper:
    """
    ..  code-block::SQL

        :x NOT IN :COLL


    :param x:
    :param coll: a sequence or a select statement.
    :return:
    """
    from .stat import SqlStat
    if not isinstance(coll, (tuple, list, SqlStat)):
        coll = [coll]
    return expr.wrap(x).not_contains(coll)


def between(x, *value) -> expr.SqlCompareOper:
    """
    https://www.sqlite.org/lang_expr.html#the_between_operator

    ..  code-block::SQL

        :x BETWEEN :value[0] AND :value[1]


    :param x:
    :param value: two value, or a range, a slice.
    :return:
    """
    return expr.wrap(x).between(*value)


def not_between(x, *value) -> expr.SqlCompareOper:
    """
    https://www.sqlite.org/lang_expr.html#the_between_operator

    ..  code-block::SQL

        :x NOT BETWEEN :value[0] AND :value[1]


    :param x:
    :param value: two value, or a range, a slice.
    :return:
    """
    return expr.wrap(x).not_between(*value)


def is_null(x) -> expr.SqlCompareOper:
    """
    ..  code-block::SQL

        :x IS NULL

    """
    return expr.wrap(x).is_null()


def is_not_null(x) -> expr.SqlCompareOper:
    """
    ..  code-block::SQL

        :x IS NOT NULL

    """
    return expr.wrap(x).is_not_null()


def excluded(t: type[T]) -> type[T]:
    """https://www.sqlite.org/lang_upsert.html"""
    return expr.SqlAlias(t, 'excluded')


@overload
def with_common_table(name: str, select: SqlSelectStat) -> expr.SqlCteExpr:
    pass


@overload
def with_common_table(name: type[T], select: SqlSelectStat) -> type[T]:
    pass


def with_common_table(name: str, select: SqlSelectStat):
    if isinstance(name, type):
        name = name.__name__
    return expr.SqlCteExpr(name, select)


# noinspection PyShadowingNames
def fields(table: type[T], *,
           primary: bool = None,
           has_default: bool = None,
           excluded: Collection[str] = None) -> tuple[expr.SqlField, ...]:
    from .table import table_fields
    fields = table_fields(table)

    if primary is not None:
        fields = [it for it in fields if it.is_primary == primary]
    if has_default is not None:
        fields = [it for it in fields if it.has_default == has_default]
    if excluded is not None:
        fields = [it for it in fields if it.name not in excluded]

    return tuple([it(table) for it in fields])



"""
Persistence Class for argp supporting
=====================================

:author:
    Ta-Shun Su

"""
import abc
from pathlib import Path
from typing import Generic, TypeVar, get_origin, get_args

from neuralib.argp import argument, copy_argument
from neuralib.persistence import (
    persistence,
    PickleHandler,
    PersistenceHandler,
    AutoIncFieldNotResolvedError
)
from neuralib.util.verbose import fprint, print_load, print_save

__all__ = ['PersistenceOptions',
           'get_options_and_cache']

T = TypeVar('T')


def persistence_filename(cache: T) -> str:
    return persistence.filename(cache) + '.pkl'


class PersistenceOptions(Generic[T], metaclass=abc.ABCMeta):
    """The Option class that handle one kind of cache class T, including
    load cache, create cache, and save cache.
    """
    GROUP_CACHE = 'Persistence options'

    invalid_cache: bool = argument(
        '--invalid-cache',
        group=GROUP_CACHE,
        help='invalid persistence data'
    )

    @property
    def persistence_class(self) -> type[T]:
        # https://stackoverflow.com/a/50101934
        for t in type(self).__orig_bases__:
            if get_origin(t) == PersistenceOptions:
                return get_args(t)[0]
        raise TypeError('unable to retrieve cache class T')

    def persistence_handler(self, dest: Path) -> PersistenceHandler[T]:
        """

        :param dest: save root directory.
        :return:
        """
        return PickleHandler(self.persistence_class, dest)

    @abc.abstractmethod
    def empty_cache(self) -> T:
        """create an empty cache which only initialize required fields.

        :return: cache instance
        """
        pass

    def find_cache(self, result: T, dest: Path = None, validator=False) -> list[T]:
        """Find the persistence.

        for all fields

        >>> template = self.empty_cache()
        >>> template.a = 1  # want to find all cache whose `a` equals to 1
        >>> template.b = field_missing # want to find all cache and don't matter what `b` is
        >>> found = self.find_cache(template)

        :param result:
        :param dest: save root directory.
        :param validator:
        :return:
        """
        handler = self.persistence_handler(dest)

        ret = []

        for file, found in handler.load_all(result):
            if validator:
                if not self.validate_cache(file, found):
                    continue

            ret.append(found)

        return ret

    def save_cache(self, result: T, dest: Path, force=True):
        """
        
        :param result:
        :param dest: save root directory.
        :param force:
        :return:
        """
        save_path = dest / persistence_filename(result)
        if save_path.exists() and not force:
            raise FileExistsError(str(save_path))

        save_path.parent.mkdir(parents=True, exist_ok=True)
        persistence.save(result, save_path)

    def load_cache(self,
                   result: T = None,
                   error_when_missing=False,
                   dest: Path = None,
                   **kwargs) -> T:
        """load persistence from disk according to *result*'s required fields.

        :param result: persistence instance with necessary fields filled.
        :param error_when_missing: do not try to generate the cache when cache missing.
        :param dest: save root directory.
        :param kwargs: overwrite field value in *result*.
        :return: persistence instance.
        :raise FileNotFoundError: error_when_missing and file not found.
        """
        do_load = not self.invalid_cache
        handler = self.persistence_handler(dest)

        if result is None:
            result = self.empty_cache()

        # load/save path
        try:
            output_file = handler.filepath(result, **kwargs)
        except AutoIncFieldNotResolvedError as e:
            if error_when_missing:
                raise FileNotFoundError from e
            do_load = False
            output_file = None

        if do_load:
            try:
                print_load(output_file)
                result = handler.load_persistence(output_file)
                if self.validate_cache(output_file, result):
                    return result
            except FileNotFoundError:
                if error_when_missing:
                    raise
            except (TypeError, ValueError, KeyError, AttributeError, RuntimeError) as e:
                fprint(repr(e), vtype='error')

        elif error_when_missing:
            raise FileNotFoundError(output_file)

        result = self.compute_cache(result)
        handler.save_persistence(result, output_file)
        output_file = handler.filepath(result)
        print_save(output_file)

        return result

    def validate_cache(self, result_path: Path, result: T) -> bool:
        """Validating loaded cache instance.

        Once validating fail (return False), goto :meth:`_compute_cache`.

        :param result_path:
        :param result:
        :return: False if validating fail.
        :raise TypeError: if validating fail.
        :raise ValueError: if validating fail.
        :raise KeyError: if validating fail.
        :raise AttributeError: if validating fail.
        :raise RuntimeError: if validating fail.
        """
        return True

    @abc.abstractmethod
    def compute_cache(self, result: T) -> T:
        """Compute cache according to *cache*'s required fields.

        :param result:
        :return: computed result
        """
        pass


def get_options_and_cache(opt_cls: type[PersistenceOptions[T]],
                          ref,
                          error_when_missing=False,
                          **kwargs) -> T:
    """
    copy the arguments from PersistenceOpt (class that compute the cache) to Apply*Opt
    Can be used for analysis apply two different cached files

     **Example**
        >>> @persistence.persistence_class
        >>> class SortIdxCache:
        ...     # for cache attributes declare
        ...     pass

        >>> class SortIdxOptions(PersistenceOptions[SortIdxCache]):
        ...     # for cache computing
        ...     def empty_cache(self) -> SortIdxCache:
        ...         pass
        ...     def _compute_cache(self, cache: SortIdxCache) -> SortIdxCache:
        ...         pass

        >>> class ApplySortIdxOptions:
        ...     # for cache applying, suppose to be the one layer before parent class for analysis class
        ...     def apply_sort_idx_cache(self) -> SortIdxCache:
        ...         return get_options_and_cache(SortIdxOptions, self)

        >>> class CPBeltSortOptions(ApplyPosBinActOptions, ApplySortIdxOptions):
        ...     pass

    :param opt_cls: PersistenceOpt
    :param ref:
    :param error_when_missing:
    :return:
    """
    return copy_argument(opt_cls(), ref).load_cache(error_when_missing=error_when_missing, **kwargs)



"""
Default Path for Cache
========================

.. code-block:: python

    CACHE_DIRECTORY = Path.home() / '.cache'
    NEUROLIB_CACHE_DIRECTORY = CACHE_DIRECTORY / 'neuralib'

    # ATLAS
    ATLAS_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'atlas'
    CCF_CACHE_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'ccf_2017'
    ALLEN_SDK_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'allensdk'
    IBL_CACHE_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'ibl'


Example of default cache under the ~/.cache/neuralib
-----------------------------------------------------
::

    neuralib
        └── atlas
            ├── allensdk
            │         ├── annotation_10.nrrd
            │         ├── manifest.json
            │         └── structures.json
            ├── ccf_2017
            │         ├── annotation_volume_10um_by_index.npy
            │         ├── structure_tree_safe_2017.csv
            │         └── template_volume_10um.npy
            ├── cellatlas.csv
            ├── cellatlas_allen_sync.csv
            └── ibl
                ├── annotation_10.nrrd
                ├── annotation_10_lut_v01.npz
                └── average_template_10.nrrd





"""
from pathlib import Path

__all__ = [
    'CACHE_DIRECTORY',
    'NEUROLIB_CACHE_DIRECTORY',
    #
    'ATLAS_CACHE_DIRECTORY',
    'CCF_CACHE_DIRECTORY',
    'ALLEN_SDK_DIRECTORY',
    'IBL_CACHE_DIRECTORY',
    #
    'CASCADE_MODEL_CACHE_DIRECTORY'
]

CACHE_DIRECTORY = Path.home() / '.cache'
NEUROLIB_CACHE_DIRECTORY = CACHE_DIRECTORY / 'neuralib'

# ATLAS
ATLAS_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'atlas'
CCF_CACHE_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'ccf_2017'
ALLEN_SDK_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'allensdk'
IBL_CACHE_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'ibl'

# MODEL
CASCADE_MODEL_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'cascade'



"""
Deeplabcut
=============

This module provide ``Deeplabcut`` result parsing
Including model meta, and output .h5 or .csv


Example of load the results
-----------------------------------

.. code-block:: python

    from neuralib.tracking.deeplabcut import *

    file = ...  # .h5 or .csv file path
    meta = ... # .pickle meta file path
    dlc = load_dlc_result(file, meta).with_global_lh_filter(lh=0.95)  # with likelihood filter >= 0.95

    # Get polars dataframe
    print(dlc.dat)

    # See all the joints
    print(dlc.joints)

    # Get a xy numpy array from specific joint. i.e., labeled Nose
    print(dlc['Nose'].xy)



Example of load the meta typeddict
-----------------------------------

.. code-block:: python

    from neuralib.tracking.deeplabcut import *

    file = ...  # .h5 or .csv file path
    meta = ... # .pickle meta file path
    dlc = load_dlc_result(file, meta).with_global_lh_filter(lh=0.95)  # with likelihood filter >= 0.95

    # See meta DeepLabCutMeta typeddict
    print(dlc.meta)

    # See model_config typeddict
    print(dlc.meta['model_config'])



"""
from .core import *



from __future__ import annotations

import abc
import math
from typing import Final, ClassVar

import attrs
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.image import AxesImage
from matplotlib.transforms import CompositeGenericTransform
from typing_extensions import Self

from neuralib.atlas.data import DATA_SOURCE_TYPE, load_ccf_annotation, load_ccf_template, load_allensdk_annotation
from neuralib.atlas.util import PLANE_TYPE, ALLEN_CCF_10um_BREGMA
from neuralib.imglib.factory import ImageProcFactory
from neuralib.typing import PathLike

__all__ = [
    'load_slice_view',
    'AbstractSliceView',
    'SlicePlane'
]


def load_slice_view(source: DATA_SOURCE_TYPE,
                    plane_type: PLANE_TYPE,
                    *,
                    output_dir: PathLike | None = None,
                    allen_annotation_res: int = 10) -> AbstractSliceView:
    """
    Load the mouse brain slice view

    :param source: ``DATA_SOURCE_TYPE``. {'ccf_annotation', 'ccf_template', 'allensdk_annotation'}
    :param plane_type: ``PLANE_TYPE``. {'coronal', 'sagittal', 'transverse'}
    :param output_dir: Output directory for caching
    :param allen_annotation_res: Volume resolution in um. default is 10 um
    :return: :class:`AbstractSliceView`
    """
    if source == 'ccf_annotation':
        data = load_ccf_annotation(output_dir)
        res = 10
    elif source == 'ccf_template':
        data = load_ccf_template(output_dir)
        res = 10
    elif source == 'allensdk_annotation':
        data = load_allensdk_annotation(resolution=allen_annotation_res, output_dir=output_dir)
        res = allen_annotation_res
    else:
        raise ValueError('')

    return AbstractSliceView(source, plane_type, res, data)


class AbstractSliceView(metaclass=abc.ABCMeta):
    """
    SliceView ABC for different `plane type`

    `Dimension parameters`:

        AP = anterior-posterior

        DV = dorsal-ventral

        ML = medial-lateral

        W = view width

        H = view height
    """
    REFERENCE_FROM: ClassVar[str] = ''
    """reference from which axis"""

    source_type: Final[DATA_SOURCE_TYPE]
    """``DATA_SOURCE_TYPE``. {'ccf_annotation', 'ccf_template', 'allensdk_annotation'}"""

    plane_type: Final[PLANE_TYPE]
    """`PLANE_TYPE``. {'coronal', 'sagittal', 'transverse'}"""

    resolution: Final[int]
    """um/pixel"""

    reference: Final[np.ndarray]
    """Array[float, [AP, DV, ML]]"""

    grid_x: Final[np.ndarray]
    """Array[int, [W, H]]"""

    grid_y: Final[np.ndarray]
    """Array[int, [W, H]]"""

    def __new__(cls, source_type: DATA_SOURCE_TYPE,
                plane: PLANE_TYPE,
                resolution: int,
                reference: np.ndarray):
        if plane == 'coronal':
            return object.__new__(CoronalSliceView)
        elif plane == 'sagittal':
            return object.__new__(SagittalSliceView)
        elif plane == 'transverse':
            return object.__new__(TransverseSliceView)
        else:
            raise ValueError(f'invalid plane: {plane}')

    def __init__(self, source_type: DATA_SOURCE_TYPE,
                 plane: PLANE_TYPE,
                 resolution: int,
                 reference: np.ndarray):
        """

        :param source_type: ``DATA_SOURCE_TYPE``. {'ccf_annotation', 'ccf_template', 'allensdk_annotation'}
        :param plane: `PLANE_TYPE``. {'coronal', 'sagittal', 'transverse'}
        :param resolution: um/pixel
        :param reference: Array[uint16, [AP, DV, ML]]
        """
        self.source_type = source_type
        self.plane_type = plane
        self.resolution = resolution
        self.reference = reference

        self.grid_y, self.grid_x = np.mgrid[0:self.height, 0:self.width]

        self._check_attrs()

    def _check_attrs(self):
        if self.resolution == 10:
            assert self.reference.shape == (1320, 800, 1140)
        elif self.resolution == 25:
            assert self.reference.shape == (528, 320, 456)

    @property
    def bregma(self) -> np.ndarray:
        if self.resolution == 10:
            return ALLEN_CCF_10um_BREGMA
        raise NotImplementedError('')

    @property
    def n_ap(self) -> int:
        """number of slices along AP axis"""
        return self.reference.shape[0]

    @property
    def n_dv(self) -> int:
        """number of slices along DV axis"""
        return self.reference.shape[1]

    @property
    def n_ml(self) -> int:
        """number of slices along ML axis"""
        return self.reference.shape[2]

    @property
    @abc.abstractmethod
    def n_planes(self) -> int:
        """number of planes in a specific plane view"""
        pass

    @property
    @abc.abstractmethod
    def width(self) -> int:
        """width (pixel) in a specific plane view"""
        pass

    @property
    @abc.abstractmethod
    def height(self) -> int:
        """height (pixel) in a specific plane view"""
        pass

    @property
    def width_mm(self) -> float:
        """width (um) in a specific plane view"""
        return self.width * self.resolution / 1000

    @property
    def height_mm(self) -> float:
        """height (um) in a specific plane view"""
        return self.height * self.resolution / 1000

    @property
    @abc.abstractmethod
    def reference_point(self) -> int:
        """reference point in a specific plane view. aka, bregma plane index"""
        pass

    @property
    @abc.abstractmethod
    def project_index(self) -> tuple[int, int, int]:
        """plane(p), x, y of index order in (AP, DV, ML)

        :return: (p, x, y)
        """
        pass

    def plane_at(self, slice_index: int) -> SlicePlane:
        return SlicePlane(slice_index, int(self.width // 2), int(self.height // 2), 0, 0, self)

    def offset(self, h: int, v: int) -> np.ndarray:
        """

        :param h: horizontal plane diff to the center. right side positive.
        :param v: vertical plane diff to the center. bottom side positive.
        :return: (H, W) array
        """
        x_frame = np.round(np.linspace(-h, h, self.width)).astype(int)
        y_frame = np.round(np.linspace(-v, v, self.height)).astype(int)
        return np.add.outer(y_frame, x_frame)

    def plane(self, offset: int | tuple[int, int, int] | np.ndarray) -> np.ndarray:
        """Get image plane.

        :param offset: Array[int, height, width] or tuple (plane, dh, dv)
        :return:
        """
        if isinstance(offset, int):
            offset = np.full_like((self.height, self.width), offset)
        elif isinstance(offset, tuple):
            offset = offset[0] + self.offset(offset[1], offset[2])
        elif not isinstance(offset, np.ndarray):
            raise TypeError(str(type(offset)))

        offset[offset < 0] = 0
        offset[offset > self.n_planes] = self.n_planes - 1

        return self.reference[self.coor_on(offset, (self.grid_x, self.grid_y))]

    def coor_on(self, plane: np.ndarray,
                o: tuple[np.ndarray, np.ndarray]) -> tuple[np.ndarray, ...]:
        """
        map slice point (x, y) at plane *plane* back to volume point (ap, dv, ml)

        :param plane: plane number of array
        :param o: tuple of (x, y)
        :return: (ap, dv, ml)
        """
        pidx, xidx, yidx = self.project_index
        ret = [0, 0, 0]
        ret[pidx] = plane
        ret[xidx] = o[0]
        ret[yidx] = o[1]

        return tuple(ret)


class CoronalSliceView(AbstractSliceView):
    REFERENCE_FROM: ClassVar[str] = 'AP'

    @property
    def n_planes(self) -> int:
        return self.n_ap

    @property
    def width(self) -> int:
        return self.n_ml

    @property
    def height(self) -> int:
        return self.n_dv

    @property
    def reference_point(self) -> int:
        return int(self.bregma[0])

    @property
    def project_index(self) -> tuple[int, int, int]:
        return 0, 2, 1


class SagittalSliceView(AbstractSliceView):
    REFERENCE_FROM: ClassVar[str] = 'ML'

    @property
    def n_planes(self) -> int:
        return self.n_ml

    @property
    def width(self) -> int:
        return self.n_ap

    @property
    def height(self) -> int:
        return self.n_dv

    @property
    def reference_point(self) -> int:
        return int(self.bregma[2])

    @property
    def project_index(self) -> tuple[int, int, int]:
        return 2, 0, 1  # p=ML, x=AP, y=DV


class TransverseSliceView(AbstractSliceView):
    REFERENCE_FROM: ClassVar[str] = 'DV'

    @property
    def n_planes(self) -> int:
        return self.n_dv

    @property
    def width(self) -> int:
        return self.n_ml

    @property
    def height(self) -> int:
        return self.n_ap

    @property
    def reference_point(self) -> int:
        return int(self.bregma[1])

    @property
    def project_index(self) -> tuple[int, int, int]:
        return 1, 2, 0


@attrs.define
class SlicePlane:
    """2D Wrapper for a specific plane"""

    slice_index: int
    """anchor index"""

    ax: int
    """anchor x"""

    ay: int
    """anchor y"""

    dw: int
    """dw in um"""

    dh: int
    """dh in um"""

    view: AbstractSliceView
    """``AbstractSliceView``"""

    @property
    def image(self) -> np.ndarray:
        return self.view.plane(self.plane_offset)

    @property
    def plane_offset(self) -> np.ndarray:
        offset = self.view.offset(self.dw, self.dh)
        return self.slice_index + offset - offset[self.ay, self.ax]

    @property
    def reference_value(self) -> float:
        """relative to reference point"""
        factor = 1000 / self.view.resolution
        return round((self.view.reference_point - self.slice_index) / factor, 2)

    def with_offset(self, dw: int, dh: int, debug: bool = False) -> Self:
        if debug:
            deg_x, deg_y = self._value_to_angle(dw, dh)
            print(f'{dw=}, {dh=}')
            print(f'{deg_x=}, {deg_y=}')

        return attrs.evolve(self, dw=dw, dh=dh)

    def _value_to_angle(self, dw: int, dh: int) -> tuple[float, float]:
        """delta value to degree"""
        rx = math.atan(2 * dw / self.view.width)
        ry = math.atan(2 * dh / self.view.height)
        deg_x = np.rad2deg(rx)
        deg_y = np.rad2deg(ry)

        return deg_x, deg_y

    def with_angle_offset(self, deg_x: float, deg_y: float) -> Self:
        """
        with degree offset

        :param deg_x: degree in x axis (width)
        :param deg_y: degree in y axis (height)
        :return:
        """
        rx = np.deg2rad(deg_x)
        ry = np.deg2rad(deg_y)

        dw = int(self.view.width * math.tan(rx) / 2)
        dh = int(self.view.height * math.tan(ry) / 2)

        return self.with_offset(dw, dh)

    def plot(self,
             ax: Axes | None = None,
             to_um: bool = True,
             with_annotation: bool = False,
             cbar: bool = False,
             with_title: bool = False,
             affine_transform: bool = False,
             customized_trans: bool = False,
             extent: tuple[float, float, float, float] | None = None,
             **kwargs) -> tuple[AxesImage, AxesImage | None, CompositeGenericTransform]:
        """
        :param ax: The Axes object on which to plot. If None, a new figure and axes are created.
        :param to_um: A boolean flag indicating whether the coordinates should be converted to micrometers. Defaults to True.
            Only applicable if ``extent`` is None.
        :param with_annotation: A boolean indicating whether to include annotations in the plot.
        :param cbar: A boolean indicating whether to include a color bar in the plot.
        :param with_title: A boolean indicating whether to include a title in the plot.
        :param affine_transform: A boolean indicating whether to apply an affine transformation to the plot.
        :param customized_trans: A boolean indicating whether to use a customized affine transformation.
        :param extent: A tuple defining the image boundaries (left, right, bottom, top). If None, boundaries are computed internally.
        :param kwargs: Additional keyword arguments passed to ``ax.imshow``.
        :return: A tuple containing the main AxesImage, the annotation AxesImage if any, and the Affine2D transformation.
        """

        if ax is None:
            _, ax = plt.subplots()

        if extent is None:
            extent = self._get_xy_range(to_um)

        #
        if affine_transform:
            import matplotlib.transforms as mtransforms
            #
            if customized_trans:
                aff = mtransforms.Affine2D(self._customized_affine_transform())
            else:
                aff = mtransforms.Affine2D().skew_deg(-20, 0)

            aff_trans = aff + ax.transData
        else:
            aff_trans = ax.transData

        #
        image = self.image.astype(float)
        image[image <= 10] = np.nan
        im_view = ax.imshow(image, cmap='Greys', extent=extent, clip_on=False, transform=aff_trans, **kwargs)

        # annotation
        if with_annotation:
            im_ann = self.plot_annotation(ax, aff_trans=aff_trans, to_um=to_um)
        else:
            im_ann = None

        #
        if cbar:
            ax.figure.colorbar(im_view)
        #
        if with_title:
            ax.set_title(f'{self.view.REFERENCE_FROM}: {self.reference_value} mm')

        ax.set(xlabel=self.unit, ylabel=self.unit)

        return im_view, im_ann, aff_trans

    unit: str = 'a.u.'

    def _get_xy_range(self, to_um: bool = True) -> tuple[float, float, float, float]:
        if to_um:
            x0 = -self.view.width_mm / 2 * 1000
            x1 = self.view.width_mm / 2 * 1000
            y0 = self.view.height_mm * 1000
            y1 = 0
            self.unit = 'um'
        else:
            x0 = -self.view.width_mm / 2
            x1 = self.view.width_mm / 2
            y0 = self.view.height
            y1 = 0
            self.unit = 'mm'

        return x0, x1, y0, y1

    def _customized_affine_transform(self) -> np.ndarray:
        # translation
        Y = np.array([[1, 0, 0], [0, 1, -4000], [0, 0, 1]])

        # shear
        tt = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [-0.03 / self.view.width, 0, 1]
        ])

        # translation
        _Y = np.array([[1, 0, 0], [0, 1, 3500], [0, 0, 1]])

        return Y @ tt @ _Y

    def plot_annotation(self,
                        ax: Axes,
                        *,
                        aff_trans: Axes.transData | None = None,
                        to_um: bool = True,
                        cmap: str = 'binary',
                        alpha: float = 0.3,
                        extent: tuple[float, float, float, float] | None = None) -> AxesImage:
        """
        Plot the annotation image

        :param ax: ``Axes``
        :param aff_trans: The transformation applied to the annotation image. If None, defaults to the Axes' transformation.
        :param to_um: A boolean flag indicating whether the coordinates should be converted to micrometers. Defaults to True.
            Only applicable if ``extent`` is None.
        :param cmap: Colormap to be used for the annotation image. Defaults to 'binary'.
        :param alpha: The imshow alpha, between 0 (transparent) and 1 (opaque). Defaults to 0.3.
        :param extent: A tuple defining the image boundaries (left, right, bottom, top). If None, boundaries are computed internally.
        :return: The AxesImage object created by imshow, representing the plotted annotation image.
        """

        if extent is None:
            extent = self._get_xy_range(to_um)

        ann_img = (
            load_slice_view('ccf_annotation', self.view.plane_type, allen_annotation_res=self.view.resolution)
            .plane(self.plane_offset)
        )

        ann = (
            ImageProcFactory(ann_img, 'RGB')
            .cvt_gray()
            .edge_detection(10, 0).image
        )

        ann = ann.astype(float)
        ann[ann <= 10] = np.nan

        if aff_trans is None:
            aff_trans = ax.transData

        im_ann = ax.imshow(ann, cmap=cmap, extent=extent, alpha=alpha, clip_on=False,
                           interpolation='none', vmin=0, vmax=255, transform=aff_trans)

        return im_ann



from __future__ import annotations

import cv2
import numpy as np

__all__ = ['affine_transform',
           'apply_transformation']


def affine_transform(src: np.ndarray) -> np.ndarray:
    """
    Do affine transformation on flatten image

    .. seealso::

        https://docs.opencv.org/4.x/d4/d61/tutorial_warp_affine.html

    :param src: source image array
    :return: output image array
    """
    srcTri = np.array([
        [0, 0],
        [src.shape[1] - 1, 0],
        [0, src.shape[0] - 1]
    ]).astype(np.float32)

    dstTri = np.array([
        [0, src.shape[1] * 0.33],
        [src.shape[1] * 0.6, src.shape[0] * 0.25],
        [src.shape[1] * 0.4, src.shape[0] * 0.7]
    ]).astype(np.float32)

    mat = cv2.getAffineTransform(srcTri, dstTri)

    warp_dst = cv2.warpAffine(src, mat, (src.shape[1], src.shape[0]))

    return warp_dst


def apply_transformation(img: np.ndarray,
                         trans_mtx: np.ndarray,
                         **kwargs) -> np.ndarray:
    """
    2D image transform

    :param img: image array
    :param trans_mtx: Transformation matrix. `Array[float, [3, 3]]`
    :param kwargs: additional arguments pass to ``cv2.warpPerspective()``
    :return:
    """
    height, width, _ = img.shape

    if trans_mtx.shape != (3, 3):
        raise ValueError(f'invalid transformation shape: {trans_mtx.shape}')

    return cv2.warpPerspective(img, trans_mtx, (width, height), **kwargs)



from pathlib import Path
from typing import ClassVar

import numpy as np
from cellpose import denoise

from neuralib.util.gpu import gpu_available
from .core import AbstractCellPoseOption, CellPoseEvalResult

__all__ = ['CellPoseAPIOption',
           'CellPoseEvalResult']


class CellPoseAPIOption(AbstractCellPoseOption):
    DESCRIPTION = 'run cell pose for directly call the api'

    RESTORE_TYPE: ClassVar[str] = 'denoise_cyto3'
    RESTORE_RATIO: ClassVar[float] = 1.0

    def run(self):
        if self.napari_view:
            self.launch_napari()
        elif self.cellpose_view:
            self.launch_cellpose_gui()
        else:
            self.eval()

    def eval(self):
        if self.file_mode:
            img = self.raw_image() if self.no_normalize else self.normalize_image()
            self._eval(self.file, img)

        elif self.batch_mode:
            iter_image = self.foreach_raw_image() if self.no_normalize else self.foreach_normalize_image()
            for name, img in iter_image:
                self._eval(name, img)

    def _eval(self, filepath: Path, image: np.ndarray):
        channel_choose = [self.chan_seg, self.chan_nuclear]
        model = denoise.CellposeDenoiseModel(
            gpu=True if gpu_available(backend='torch') else False,
            model_type=self.model,
            restore_type=self.RESTORE_TYPE,
            chan2_restore=True
        )

        masks, flows, styles, imgs_dn = model.eval(
            image,
            channels=channel_choose,
            diameter=self.diameter
        )

        ret = CellPoseEvalResult(
            filepath.name,
            image,
            self.diameter,
            channel_choose,
            masks,
            flows,
            styles,
            img_restore=imgs_dn,
            restore=self.RESTORE_TYPE,
            ratio=self.RESTORE_RATIO
        )

        ret.save_seg_file(str(filepath))

        if self.save_ij_roi:
            ret.save_roi(self.ij_roi_output(filepath))


if __name__ == '__main__':
    CellPoseAPIOption().main()



from __future__ import annotations

import datetime
from pathlib import Path
from typing import Literal, TypedDict, final

import attrs
import numpy as np
import polars as pl

from neuralib.calimg.cellular import CellularCoordinates
from neuralib.typing import PathLike
from neuralib.util.verbose import fprint

__all__ = [
    'SIGNAL_TYPE',
    'CALCIUM_TYPE',
    #
    'Suite2pGUIOptions',
    'Suite2pRoiStat',
    'Suite2PResult',
    #
    'get_s2p_coords'
]

SIGNAL_TYPE = Literal['df_f', 'spks']
CALCIUM_TYPE = Literal['baseline', 'transient']


class Suite2pGUIOptions(TypedDict, total=False):
    """ Suite2p GUI setting.

    .. seealso:: `<https://suite2p.readthedocs.io/en/latest/settings.html>`_"""
    look_one_level_down: float
    fast_disk: str
    delete_bin: bool
    mesoscan: bool
    bruker: bool
    h5py: list
    h5py_key: str
    save_path0: str
    save_folder: str
    subfolders: list
    move_bin: bool
    nplanes: int
    nchannels: int
    functional_chan: int
    tau: float
    fs: float
    force_sktiff: bool
    frames_include: int
    multiplane_parallel: float
    preclassify: float
    save_mat: bool
    save_NWB: float
    combined: float
    aspect: float
    do_bidiphase: bool
    bidiphase: float
    bidi_corrected: bool
    do_registration: int
    two_step_registration: float
    keep_movie_raw: bool
    nimg_init: int
    batch_size: int
    maxregshift: float
    align_by_chan: int
    reg_tif: bool
    reg_tif_chan2: bool
    subpixel: int
    smooth_sigma_time: float
    smooth_sigma: float
    th_badframes: float
    norm_frames: bool
    force_refImg: bool
    pad_fft: bool
    nonrigid: bool
    block_size: tuple[int, int]
    snr_thresh: float
    maxregshiftNR: float
    oneP_reg: bool  # 1Preg
    spatial_hp: int
    spatial_hp_reg: float
    spatial_hp_detect: int
    pre_smooth: float
    spatial_taper: float
    roidetect: bool
    spikedetect: bool
    anatomical_only: float
    sparse_mode: bool
    diameter: float
    spatial_scale: float
    connected: bool
    nbinned: int
    max_iterations: int
    threshold_scaling: float
    max_overlap: float
    high_pass: float
    denoise: bool
    soma_crop: bool
    neuropil_extract: bool
    inner_neuropil_radius: float
    min_neuropil_pixels: int
    lam_percentile: float
    allow_overlap: bool
    use_builtin_classifier: bool
    classifier_path: int
    chan2_thres: float
    baseline: str
    win_baseline: float
    sig_baseline: float
    prctile_baseline: float
    neucoeff: int
    suite2p_version: str
    data_path: list[str]
    sbx_ndeadcols: int
    input_format: str
    save_path: str
    ops_path: str
    reg_file: str
    filelist: list[str]
    nframes_per_folder: np.ndarray
    sbx_ndeadrows: int
    meanImg: np.ndarray
    meanImg_chan2: np.ndarray  # if chan_2
    nframes: int
    Ly: int
    Lx: int
    date_proc: datetime.datetime
    refImg: np.ndarray
    rmin: int
    rmax: int
    yblock: list[np.ndarray]
    xblock: list[np.ndarray]
    nblocks: list[int]
    NRsm: np.ndarray
    yoff: np.ndarray
    xoff: np.ndarray
    corrXY: np.ndarray
    yoff1: np.ndarray
    xoff1: np.ndarray
    corrXY1: np.ndarray
    badframes: np.ndarray
    yrange: list[int]
    xrange: list[int]
    tPC: np.ndarray
    regPC: np.ndarray
    regDX: np.ndarray
    Lyc: int
    Lxc: int
    max_proj: np.ndarray
    Vmax: np.ndarray
    ihop: np.ndarray
    Vsplit: np.ndarray
    Vcorr: np.ndarray
    Vmap: list[np.ndarray]
    spatscale_pix: np.ndarray
    meanImgE: np.ndarray
    timing: dict[str, float]


class Suite2pRoiStat(TypedDict, total=False):
    """Suite2p GUI imaging.

    .. seealso:: `<https://suite2p.readthedocs.io/en/latest/outputs.html#stat-npy-fields>`_
    """
    ypix: np.ndarray
    xpix: np.ndarray
    lam: np.ndarray
    med: list[int, int]
    footprint: float
    mrs: float
    mrs0: float
    compact: float
    solidity: float
    npix: int
    npix_soma: int
    soma_crop: np.ndarray
    overlap: np.ndarray
    radius: float
    aspect_ratio: float
    npix_norm_no_crop: float
    npix_norm: float
    skew: float
    std: float


# ============== #
# Suite2P Result #
# ============== #

@final
@attrs.frozen
class Suite2PResult:
    """suite2p result container

    `Dimension parameters`:

        N: number of neurons

        F: number pf frames

        W: image width

        H: image height
    """
    directory: Path
    """Directory contain all the s2p output files"""

    f_raw: np.ndarray
    """Fluorescence traces 2D array. `Array[float, [N, F]]`"""

    f_neu: np.ndarray
    """Neuropil fluorescence traces 2D array. `Array[float, [N, F]]`"""

    spks: np.ndarray
    """Deconvolved activity 2D array. `Array[float, [N, F]]`"""

    stat: np.ndarray
    """GUI imaging after registration, i.e., x, ypixel., etc. `Array[Suite2pRoiStat, N]`"""

    ops: Suite2pGUIOptions
    """GUI options"""

    iscell: np.ndarray
    """Cell probability for each ROI. `Array[float, [N, 2]]`"""

    cell_prob_thres: float | None
    """Cell probability threshold for loading the data"""

    redcell: np.ndarray | None = attrs.field(default=None)
    """Red cell probability 2D array. `Array[float, [N, 2]]`"""

    redcell_threshold: float | None = attrs.field(default=None)
    """Red cell probability threshold"""

    runtime_frate_check: float | None = attrs.field(default=None)
    """If not None, check frame rate lower bound"""

    def __attrs_post_init__(self):
        if self.runtime_frate_check is not None:
            self._check_frame_rate()

    def _check_frame_rate(self):
        """User specific check suite2p config is correct"""
        if self.fs * self.n_plane > self.runtime_frate_check:
            fprint(f'the fr: {self.fs} and n_etl: {self.n_plane} might not set properly in suite2p,'
                   f'check output ops.json', vtype='error')
            raise RuntimeError('fs and n_plane are not set properly in suite2p')

    @classmethod
    def launch_gui(cls, directory: PathLike) -> None:
        """
        launch the suite2p GUI

        :param directory: directory contain all the s2p output files. e.g., <SUITE2P_OUTPUT>/suite2p/plane<P>
        :return:
        """
        from suite2p.gui import gui2p

        if not isinstance(directory, Path):
            directory = Path(directory)
        gui2p.run(str(directory / 'stat.npy'))

    @classmethod
    def load(
            cls,
            directory: PathLike,
            cell_prob_thres: float | None = 0.5,
            red_cell_threshold: float = 0.65,
            channel: int = 0,
            runconfig_frate: float | None = 30.0,
    ) -> Suite2PResult:
        """
        Load suite2p result from directory

        :param directory: Directory contain all the s2p output files. e.g., */suite2p/plane[P]
        :param cell_prob_thres: Cell probability. If float type, mask for the value in ``iscell[:, 1]``.
                    If None, use the binary criteria in GUI output
        :param red_cell_threshold: Red cell threshold
        :param channel: channel (PMT) Number for the functional channel. i.e., 0 if GCaMP, 1 if jRGECO in scanbox setting
        :param runconfig_frate: if not None, check frame rate lower-bound to make sure the s2p runconfig
        :return: :class:`Suite2PResult`
        """
        if not isinstance(directory, Path):
            directory = Path(directory)

        if channel == 0:
            F = np.load(directory / 'F.npy', allow_pickle=True)
            FNeu = np.load(directory / 'Fneu.npy', allow_pickle=True)
            spks = np.load(directory / 'spks.npy')
            stat = np.load(directory / 'stat.npy', allow_pickle=True)
            ops = np.load(directory / 'ops.npy', allow_pickle=True).tolist()
            iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
            r = directory / 'redcell.npy'
            if r.exists():
                redcell = np.load(r, allow_pickle=True)
            else:
                redcell = None

        elif channel == 1:  # second channel
            F = np.load(directory / 'F_chan2.npy', allow_pickle=True)
            FNeu = np.load(directory / 'Fneu_chan2.npy', allow_pickle=True)
            spks = np.load(directory / 'spks.npy')
            stat = np.load(directory / 'stat.npy', allow_pickle=True)
            ops = np.load(directory / 'ops.npy', allow_pickle=True).tolist()
            iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
            redcell = None
        else:
            raise IndexError(f'{channel} unknown')

        #
        if cell_prob_thres is None:
            x = iscell[:, 0] == 1
        elif isinstance(cell_prob_thres, float):
            x = iscell[:, 1] >= cell_prob_thres
        else:
            raise TypeError(f'invalid type: {type(cell_prob_thres)}')

        F = F[x]
        FNeu = FNeu[x]
        spks = spks[x]
        stat = stat[x]
        iscell = iscell[x]
        if redcell is not None:
            redcell = redcell[x]

        return Suite2PResult(
            directory,
            F,
            FNeu,
            spks,
            stat,
            ops,
            iscell,
            cell_prob_thres,
            redcell,
            red_cell_threshold if channel == 1 else None,
            runconfig_frate
        )

    @property
    def has_chan2(self) -> bool:
        """if has a second channel"""
        if self.ops['nchannels'] == 2:
            return True
        return False

    @property
    def n_neurons(self) -> int:
        """number of neurons after load.
        could be less than GUI ROI number if use higher cell_prob in
        :meth:`~neuralib.calimg.suite2p.core.Suite2PResult.load()`"""
        return self.f_raw.shape[0]

    @property
    def n_frame(self) -> int:
        """number of frame number"""
        return self.f_raw.shape[1]

    @property
    def cell_prob(self) -> np.ndarray:
        """probability that the ROI is a cell based on the default classifier. `Array[float, N]`"""
        return self.iscell[:, 1]

    @property
    def n_red_neuron(self) -> int:
        """number of identified neuron based on red cell threshold"""
        if self.has_chan2:
            return np.count_nonzero(self.red_cell_prob >= self.redcell_threshold)
        else:
            raise RuntimeError('no channel 2')

    @property
    def red_cell_prob(self) -> np.ndarray | None:
        """red cell probability, `Array[float, N]`"""
        if self.redcell is None:
            return None
        return self.redcell[:, 1]

    @property
    def signal_baseline(self) -> float:
        """Gaussian filter width in seconds"""
        return self.ops['sig_baseline']

    @property
    def window_baseline(self) -> float:
        """window for max/min filter in seconds"""
        return self.ops['win_baseline']

    @property
    def fs(self) -> float:
        """suite2p approximate frame rate per plane, exact value should be checked in .sbx or .mat"""
        return self.ops['fs']

    @property
    def neucoeff(self) -> float:
        """neuropil coefficient, normally should be ~0.7"""
        return self.ops['neucoeff']

    @property
    def prctile_baseline(self) -> float:
        """percentile of trace to use as baseline if ops['baseline'] = constant_percentile"""
        return self.ops['prctile_baseline']

    @property
    def n_plane(self) -> int:
        """number of optical plane"""
        return self.ops['nplanes']

    @property
    def image_width(self) -> int:
        """image width (in pixel)"""
        return self.ops['Lx']

    @property
    def image_height(self) -> int:
        """image height (in pixel)"""
        return self.ops['Ly']

    @property
    def image_mean(self) -> np.ndarray:
        """mean image for chan0(1st). `Array[float, [H, W]]`"""
        return self.ops['meanImg'].T

    @property
    def image_mean_ch2(self) -> np.ndarray:
        """mean image for chan1(2nd). `Array[float, [H, W]]`"""
        return self.ops['meanImg_chan2'].T

    @property
    def indicator_tau(self) -> float:
        """The timescale of the sensor (in seconds)"""
        return self.ops['tau']

    @property
    def rigid_x_offsets(self) -> np.ndarray:
        """x-shifts of recording at each timepoint. `Array[int, F]`"""
        return self.ops['xoff']

    @property
    def rigid_y_offsets(self) -> np.ndarray:
        """y-shifts of recording at each timepoint. `Array[int, F]`"""
        return self.ops['yoff']

    @property
    def rigid_xy_offset(self) -> np.ndarray:
        """peak of phase correlation between frame and reference image at each timepoint. `Array[float, F]`"""
        return self.ops['corrXY']

    @property
    def nonrigid_x_offsets(self) -> np.ndarray:
        """(frames, block_size). `Array[float, F]`"""
        return self.ops['xoff1']

    @property
    def nonrigid_y_offsets(self) -> np.ndarray:
        """`Array[float, F]`"""
        return self.ops['yoff1']

    @property
    def nonrigid_xy_offsets(self) -> np.ndarray:
        """`Array[float, F]`"""
        return self.ops['corrXY1']

    @classmethod
    def load_total_neuron_number(cls,
                                 directory: Path,
                                 cell_prob: float | None = 0.5) -> int:
        """
        Load number of neuron based on iscell.npy

        :param directory: directory contains the iscell.npy
        :param cell_prob: cell probability,
                    bool type: use the binary criteria in GUI output
                    float type: value in ``iscell[:, 1]``
        :return: Number of neurons
        """
        iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
        if cell_prob is None:
            return np.count_nonzero(iscell[:, 0] == 1)
        else:
            return np.count_nonzero(iscell[:, 1] >= cell_prob)

    def get_rois_pixels(self) -> np.ndarray:
        """ROIs pixel (N, 2)"""
        ret = np.zeros((self.n_neurons, 2))
        for i in range(self.n_neurons):
            x = np.mean(self.stat[i]['xpix'])
            y = np.mean(self.stat[i]['ypix'])
            ret[i] = np.array([x, y])

        return ret

    def get_neuron_id_mapping(self) -> pl.DataFrame:
        """
        Retrieves a mapping between neuron IDs and their corresponding raw indices
        based on whether the cell detection probabilities meet a specified threshold.
        If no cell detection probabilities are provided, the mapping assumes all
        indices are valid neurons.

        :return: A Polars DataFrame containing two columns: `neuron_id` and  `raw_index`.
        """
        n = np.arange(len(self.f_raw))
        if self.cell_prob is None:
            return pl.DataFrame([n, n], schema=['neuron_id', 'raw_index'], orient='col')
        else:
            iscell = np.load(self.directory / 'iscell.npy', allow_pickle=True)
            mx = np.nonzero(iscell[:, 1] >= self.cell_prob)[0]
            return pl.DataFrame([n, mx], schema=['neuron_id', 'raw_index'], orient='col')


def get_s2p_coords(s2p: Suite2PResult,
                   neuron_list: int | list[int] | slice | np.ndarray | None,
                   plane_index: int,
                   factor: float) -> CellularCoordinates:
    """
    Get the suite2p coordinates of all cells.

    :param s2p: ``Suite2PResult``
    :param neuron_list: neuron index or index list/arr. If None, then load all neurons
    :param plane_index: optic plane index
    :param factor: pixel to mm factor
    :return: :class:`~neuralib.calimg.cellular_cords.CellularCoordinates`
    """
    if neuron_list is None:
        neuron_list = np.arange(s2p.n_neurons)

    n_neurons = len(neuron_list)
    xpix = np.zeros(n_neurons)
    ypix = np.zeros(n_neurons)

    for i, n in enumerate(neuron_list):
        xpix[i] = np.mean(s2p.stat[i]['xpix'])
        ypix[i] = np.mean(s2p.stat[i]['ypix'])

    xcord = xpix * factor / 1000  # ap
    ycord = ypix * factor / 1000  # ml

    if isinstance(neuron_list, int):
        src_plane_index = plane_index
    elif isinstance(neuron_list, (list, np.ndarray, slice)):
        src_plane_index = np.full_like(neuron_list, plane_index)
    else:
        raise TypeError('')

    return CellularCoordinates(
        np.array(neuron_list),
        xcord,
        ycord,
        plane_index=plane_index,
        unit='mm',
        source_plane_index=src_plane_index
    )



from __future__ import annotations

import dataclasses
from functools import cached_property
from pathlib import Path
from typing import final, NamedTuple, Literal, Any
from xml.etree.ElementTree import tostring

import aicspylibczi
import attrs
import matplotlib.pyplot as plt
import numpy as np
import xmltodict
from matplotlib.axes import Axes

from neuralib.scanner import AbstractConfocalScanner, SceneIdx, DimCode, ZEISS_CZI_CHANNELS_ORDER, ZPROJ_TYPE
from neuralib.typing import PathLike
from neuralib.util.utils import joinn

__all__ = ['CziConfocalScanner']


class ImageFileInfo(NamedTuple):
    channel: int | Literal['merge']
    stack: int
    tile: int

    scene: int | None = None

    def build_filename(self, *arg, sep: str = '_') -> str:
        ret = joinn(sep, *arg)
        if self.scene is not None:
            ret += f'_s{self.scene}'
        ret += f'_c{self.channel}_z{self.stack}_t{self.tile}.tiff'
        return ret

    @property
    def channel_name(self) -> str:
        if isinstance(self.channel, int):
            return ZEISS_CZI_CHANNELS_ORDER[self.channel]
        else:
            return self.channel


# TODO sep plot func from the class
@final
@attrs.define
class CziConfocalScanner(AbstractConfocalScanner):
    """czi confocal image data"""

    czifile: aicspylibczi.CziFile
    """`aicspylibczi.CziFile`"""
    filepath: Path
    meta: dict[str, Any]
    n_scenes: int = dataclasses.field(init=False, default=1)
    consistent_scan_configs: bool = dataclasses.field(init=False)
    """whether same configs (i.e., X, Y, C ...) in different scenes"""

    def __attrs_post_init__(self):
        czi_shape = self.czifile.get_dims_shape()
        s1 = czi_shape[0]['S'][1]

        if len(czi_shape) == 1 and s1 == 1:
            self.n_scenes = 1
            self.consistent_scan_configs = True

        elif len(czi_shape) == 1 and s1 != 1:
            self.n_scenes = s1
            self.consistent_scan_configs = True
        else:
            self.n_scenes = len(czi_shape)
            self.consistent_scan_configs = False

    @classmethod
    def load(cls, file: PathLike) -> CziConfocalScanner:
        czi_file = aicspylibczi.CziFile(file)
        xml = tostring(czi_file.meta, encoding='utf-8').decode('utf-8')
        return CziConfocalScanner(
            czi_file,
            Path(file),
            xmltodict.parse(xml)
        )

    @cached_property
    def tile_info(self) -> dict[str, Any]:
        return self.meta['ImageDocument']['Metadata']['Experiment']['ExperimentBlocks']['AcquisitionBlock'][
            'SubDimensionSetups']['RegionsSetup']['SampleHolder']['TileRegions']['TileRegion']

    @property
    def tile_ncols(self) -> int:
        """how many tiles for each column"""
        return self.tile_info['Columns']

    @property
    def tile_nrows(self) -> int:
        """how many tiles for each row"""
        return self.tile_info['Rows']

    def get_dim_code(self) -> DimCode:
        return DimCode(self.czifile.dims)

    @property
    def width(self) -> dict[SceneIdx, int]:
        return {s: self._get_configs('X')[s] for s in range(self.n_scenes)}

    @property
    def height(self) -> dict[SceneIdx, int]:
        return {s: self._get_configs('Y')[s] for s in range(self.n_scenes)}

    @property
    def n_phases(self) -> dict[SceneIdx, int]:
        """how many scanning face"""
        return {s: self._get_configs('H')[s] for s in range(self.n_scenes)}

    @property
    def n_channels(self) -> dict[SceneIdx, int]:
        return {s: self._get_configs('C')[s] for s in range(self.n_scenes)}

    @property
    def n_zstacks(self) -> dict[SceneIdx, int]:
        return {s: self._get_configs('Z')[s] for s in range(self.n_scenes)}

    @property
    def n_tiles(self) -> dict[SceneIdx, int]:
        return {s: self._get_configs('M')[s] for s in range(self.n_scenes)}

    @property
    def is_mosaic(self) -> bool:
        return any(value > 1 for value in self.n_tiles.values())

    def _get_configs(self, code: str) -> list[int]:
        """get scanning configs"""
        dims = self.get_dim_code()

        if code not in dims:
            raise ValueError(f'unknown code: {code}')

        if self.consistent_scan_configs:
            size = self.czifile.size
            return [size[dims.index(code)]] * self.n_scenes
        else:
            size = [it[code] for it in self.czifile.get_dims_shape()]
            return [it[1] for it in size]

    def get_image(self,
                  channel: int,
                  scene: SceneIdx | None = None,
                  depth: int | slice | np.ndarray | None = None,
                  zproj_type: ZPROJ_TYPE = 'max',
                  norm: bool = True) -> np.ndarray:
        """
        Get the image array

        :param channel: channel index
        :param scene: scanning position
        :param depth: z stacks index, if None, use all stacks
        :param zproj_type: which z projection type, refer to fiji
        :param norm: normalization, for visualization
        :return: (Y, X)
        """

        kwargs = {}
        if self.is_mosaic:
            fn = self.czifile.read_mosaic  # tiles
        else:
            fn = self.czifile.read_image
            # TODO prob better way, seems S kw only available in non-mosaic image
            scene = 0 if scene is None else scene
            kwargs.update({'S': scene})

        #
        if depth is None:
            depth = np.arange(self.n_zstacks[scene])
        if isinstance(depth, int):
            ret = fn(C=channel, Z=depth, **kwargs).squeeze()
        elif isinstance(depth, (np.ndarray, slice)):
            stacks = np.array([fn(C=channel, Z=z, **kwargs)[0].squeeze() for z in depth])  # (C, )
            ret = self._zproj(stacks, zproj_type)
        else:
            raise TypeError(f'unknown type {type(depth)}')

        if norm:
            ret = ret / np.max(ret)

        return ret

    def get_coordinates(self) -> list[tuple[int, int]]:
        """(P, 2) with xy coordinates"""
        return [(it.x, it.y) for it in self.czifile.get_all_scene_bounding_boxes().values()]

    def imshow(self,
               scene: SceneIdx | None = None,
               add_scale_bar: bool = True,
               output: PathLike | None = None,
               position_only=False):
        """Simple plot for specific config"""
        if self.n_scenes > 1:
            self._imshow_multiple_scene(scene, position_only)
        else:
            self._imshow_single_scene(scene)

        if add_scale_bar:
            # TODO
            pass

        if output is not None:
            plt.savefig(output)
        else:
            plt.show()

    def _imshow_single_scene(self, scene: SceneIdx | None):
        fig, ax = plt.subplots()
        self._plot_merge_channel(ax, scene)

    def _imshow_multiple_scene(self, scene: SceneIdx | None,
                               position_only=False, **kwargs):
        fig, ax = plt.subplots(1, 2, figsize=(16, 4))

        if not position_only:
            self._plot_merge_channel(ax[0], scene, **kwargs)

        xy = self.get_coordinates()
        for i, (x, y) in enumerate(xy):
            ax[1].plot(x, y)
            ax[1].text(x, y, s=i)
        ax[1].invert_yaxis()

    def _plot_merge_channel(self, ax: Axes,
                            scene: SceneIdx | None,
                            **kwargs):

        scene = 0 if scene is None else scene

        for c in range(self.n_channels[scene]):
            img = self.get_image(c, scene, norm=True)
            ax.imshow(img, cmap=ZEISS_CZI_CHANNELS_ORDER[c], alpha=0.8, vmax=0.1, aspect='auto', **kwargs)

    def foreach_tif_output(self, output: PathLike | None = None,
                           combine_channels: bool = False,
                           combine_tiles: bool = True):
        """

        :param output: output directory
        :param combine_channels:
        :param combine_tiles: TODO overlap compensation?
        :return:
        """

        if output is None:
            output = self.filepath.parent / self.filepath.stem
            output.mkdir(exist_ok=True, parents=True)

        if self.is_mosaic:
            # HSTCZMYX
            images = self.czifile.read_image()[0]  # (1, 1, 1, 3, 10, 112, 512, 512)
            images = np.squeeze(images)  # (3, 10, 112, 512, 512)

            if combine_channels:
                self._foreach_output_merge_channel(images, output)
            else:
                self._foreach_output_channel_sep(images, output)

        else:
            raise NotImplementedError('')

    def _foreach_output_channel_sep(self, images: np.ndarray, output: Path):
        import tifffile

        for i, ch in enumerate(images):
            for j, z in enumerate(ch):
                for k, t in enumerate(z):
                    file = ImageFileInfo(i, j, k).build_filename(*self.filepath.stem.split('_'))

                    tifffile.imwrite(output / file, t)

    def _foreach_output_merge_channel(self, images: np.ndarray, output: Path):
        """TODO (Y, X) to (Y, X, 3)?"""
        pass


def main():
    import argparse

    ap = argparse.ArgumentParser()

    ap.add_argument('-F', '--czi',
                    metavar='FILE',
                    type=Path,
                    help='czi_file filepath',
                    dest='file')

    ap.add_argument('-O', '--output',
                    metavar='FILE',
                    type=Path,
                    default=None,
                    help='output filepath',
                    dest='output')

    ap.add_argument('-P', '--position',
                    action='store_true',
                    help='only show position fig for speeding up loading image',
                    dest='position')

    ap.add_argument('-E', '--export',
                    action='store_true',
                    help='foreach export as tiff files',
                    dest='export')

    opt = ap.parse_args()

    czi = CziConfocalScanner.load(opt.file)
    if opt.export:
        czi.foreach_tif_output(output=opt.output)
    else:
        czi.imshow(output=opt.output, position_only=opt.position)


if __name__ == '__main__':
    main()



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Literal

import nrrd
import numpy as np
import pandas as pd
import polars as pl
from brainglobe_atlasapi import BrainGlobeAtlas

from neuralib.io.core import CCF_CACHE_DIRECTORY, ALLEN_SDK_DIRECTORY
from neuralib.typing import PathLike
from neuralib.util.tqdm import download_with_tqdm
from neuralib.util.verbose import fprint

__all__ = [
    'DATA_SOURCE_TYPE',
    #
    'load_ccf_annotation',
    'load_ccf_template',
    'load_structure_tree',
    #
    'load_allensdk_annotation',
    #
    'load_bg_structure_tree'

]

DATA_SOURCE_TYPE = Literal['ccf_annotation', 'ccf_template', 'allensdk_annotation']


# ===================== #
# AllenCCF Data Source #
# ===================== #

def _cache_nparray(url: str, file: PathLike) -> None:
    fprint(f'DOWNLOADING... {file.name} from {url}', vtype='io')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        content = download_with_tqdm(url)
        data = np.load(content, allow_pickle=True)
        np.save(file, data)


def load_ccf_annotation(output_dir: PathLike | None = None) -> np.ndarray:
    """
    Annotation volume file from AllenCCF pipeline

    .. seealso::

        https://github.com/cortex-lab/allenCCF

    :param output_dir: output directory for caching
    :return: Array[uint16, [AP, DV, ML]]
    """
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / 'annotation_volume_10um_by_index.npy'

    if not file.exists():
        url = 'https://figshare.com/ndownloader/files/44925493'
        _cache_nparray(url, file)

    return np.load(file, allow_pickle=True)


def load_ccf_template(output_dir: PathLike | None = None) -> np.ndarray:
    """
    Template volume file from AllenCCF pipeline

    .. seealso::

        https://github.com/cortex-lab/allenCCF

    :param output_dir: output directory for caching
    :return: Array[uint16, [AP, DV, ML]]
    """
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / 'template_volume_10um.npy'

    if not file.exists():
        url = 'https://figshare.com/ndownloader/files/44925496'
        _cache_nparray(url, file)

    return np.load(file, allow_pickle=True)


def load_structure_tree(version: Literal['2017', 'old'] = '2017',
                        output_dir: PathLike | None = None) -> pl.DataFrame:
    """
    Load structure tree dataframe

    :param version: {'2017', 'old'}
    :param output_dir: output directory for caching
    :return: structure tree dataframe

    ::

        ┌───────────┬──────────┬───────────────────────────────┬─────────┬──────────┬─────────────┬───────────────┬────────┬─────────────────────┬───────┬──────────┬─────────────┬───────────────────┬───────────────────┬─────────────────────────┬──────────────────────────────┬────────┬───────────┬──────────────────────┬──────────────┬───────────────────────────────┐
        │ id        ┆ atlas_id ┆ name                          ┆ acronym ┆ st_level ┆ ontology_id ┆ hemisphere_id ┆ weight ┆ parent_structure_id ┆ depth ┆ graph_id ┆ graph_order ┆ structure_id_path ┆ color_hex_triplet ┆ neuro_name_structure_id ┆ neuro_name_structure_id_path ┆ failed ┆ sphinx_id ┆ structure_name_facet ┆ failed_facet ┆ safe_name                     │
        │ ---       ┆ ---      ┆ ---                           ┆ ---     ┆ ---      ┆ ---         ┆ ---           ┆ ---    ┆ ---                 ┆ ---   ┆ ---      ┆ ---         ┆ ---               ┆ ---               ┆ ---                     ┆ ---                          ┆ ---    ┆ ---       ┆ ---                  ┆ ---          ┆ ---                           │
        │ i64       ┆ f64      ┆ str                           ┆ str     ┆ str      ┆ i64         ┆ i64           ┆ i64    ┆ f64                 ┆ i64   ┆ i64      ┆ i64         ┆ list[i64]         ┆ str               ┆ str                     ┆ str                          ┆ str    ┆ i64       ┆ i64                  ┆ i64          ┆ str                           │
        ╞═══════════╪══════════╪═══════════════════════════════╪═════════╪══════════╪═════════════╪═══════════════╪════════╪═════════════════════╪═══════╪══════════╪═════════════╪═══════════════════╪═══════════════════╪═════════════════════════╪══════════════════════════════╪════════╪═══════════╪══════════════════════╪══════════════╪═══════════════════════════════╡
        │ 997       ┆ -1.0     ┆ root                          ┆ root    ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ -1.0                ┆ 0     ┆ 1        ┆ 0           ┆ [997]             ┆ FFFFFF            ┆ null                    ┆ null                         ┆ f      ┆ 1         ┆ 385153371            ┆ 734881840    ┆ root                          │
        │ 8         ┆ 0.0      ┆ Basic cell groups and regions ┆ grey    ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 997.0               ┆ 1     ┆ 1        ┆ 1           ┆ [997, 8]          ┆ BFDAE3            ┆ null                    ┆ null                         ┆ f      ┆ 2         ┆ 2244697386           ┆ 734881840    ┆ Basic cell groups and regions │
        │ …         ┆ …        ┆ …                             ┆ …       ┆ …        ┆ …           ┆ …             ┆ …      ┆ …                   ┆ …     ┆ …        ┆ …           ┆ …                 ┆ …                 ┆ …                       ┆ …                            ┆ …      ┆ …         ┆ …                    ┆ …            ┆ …                             │
        │ 65        ┆ 715.0    ┆ parafloccular sulcus          ┆ pfs     ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 1040.0              ┆ 3     ┆ 1        ┆ 1324        ┆ [997, 1024, … 65] ┆ AAAAAA            ┆ null                    ┆ null                         ┆ f      ┆ 1325      ┆ 771629690            ┆ 734881840    ┆ parafloccular sulcus          │
        │ 624       ┆ 926.0    ┆ Interpeduncular fossa         ┆ IPF     ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 1024.0              ┆ 2     ┆ 1        ┆ 1325        ┆ [997, 1024, 624]  ┆ AAAAAA            ┆ null                    ┆ null                         ┆ f      ┆ 1326      ┆ 1476705011           ┆ 734881840    ┆ Interpeduncular fossa         │
        │ 304325711 ┆ -2.0     ┆ retina                        ┆ retina  ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 997.0               ┆ 1     ┆ 1        ┆ 1326        ┆ [997, 304325711]  ┆ 7F2E7E            ┆ null                    ┆ null                         ┆ f      ┆ 1327      ┆ 3295290839           ┆ 734881840    ┆ retina                        │
        └───────────┴──────────┴───────────────────────────────┴─────────┴──────────┴─────────────┴───────────────┴────────┴─────────────────────┴───────┴──────────┴─────────────┴───────────────────┴───────────────────┴─────────────────────────┴──────────────────────────────┴────────┴───────────┴──────────────────────┴──────────────┴───────────────────────────────┘

    """
    #
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)
    #
    if version == '2017':
        filename = 'structure_tree_safe_2017.csv'
    elif version == 'old':
        filename = 'structure_tree_safe.csv'
    else:
        raise ValueError('')

    #
    file = output_dir / filename
    if not file.exists():
        url = f'https://raw.githubusercontent.com/cortex-lab/allenCCF/master/{filename}'
        content = download_with_tqdm(url)
        df = pd.read_csv(content)
        df.to_csv(file)
        fprint(f'DOWNLOAD! {filename} in {output_dir}', vtype='io')

    ret = (pl.read_csv(file)
           .with_columns(pl.col('parent_structure_id').fill_null(-1))
           .with_columns(pl.col('structure_id_path')
                         .map_elements(lambda it: tuple(map(int, it[1:-1].split('/'))),
                                       return_dtype=pl.List(pl.Int64))))
    #
    if '2017' in Path(file).name:
        ret = ret.with_columns(pl.col('atlas_id').fill_null(-2))
    else:
        ret = ret.with_columns(pl.col('name').alias('safe_name'))

    return ret


# ===================== #
# Allen SDK Data Source #
# ===================== #

def load_allensdk_annotation(resolution: int = 10, output_dir: PathLike | None = None) -> np.ndarray:
    """
    Data Source directly from Allen Institute

    .. seealso::

        https://download.alleninstitute.org/informatics-archive/current-release/mouse_ccf/annotation/

    :param resolution: volume resolution in um. default is 10 um
    :param output_dir: output directory for caching
    :return: Array[uint32, [AP, DV, ML]]
    """
    if output_dir is None:
        output_dir = ALLEN_SDK_DIRECTORY
        if not ALLEN_SDK_DIRECTORY.exists():
            ALLEN_SDK_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / f'annotation_{resolution}.nrrd'

    if not file.exists():
        try:
            from allensdk.api.queries.mouse_connectivity_api import MouseConnectivityApi
        except ImportError as e:
            fprint(
                'Build error from project.toml. Please manually install using "pip install allensdk --no-deps"',
                vtype='error'
            )
            raise e

        mcapi = MouseConnectivityApi()
        version = MouseConnectivityApi.CCF_VERSION_DEFAULT

        mcapi.download_annotation_volume(version, resolution, file)

    return nrrd.read(file)[0]


# =========================== #
# BrainGlobeAtlas Data Source #
# =========================== #

def load_bg_structure_tree(source: str = 'allen_mouse_10um',
                           check_latest: bool = True,
                           parse: bool = False) -> pl.DataFrame:
    """
    Load structure dataframe or dict from `brainglobe_atlasapi`

    :param source: allen source name
    :param check_latest: if check the brainglobe api latest version
    :param parse: whether parse the child and parent in the same row
    :return:
    """
    file = BrainGlobeAtlas(source, check_latest=check_latest).root_dir / 'structures.csv'
    df = pl.read_csv(file)

    if parse:
        name = df.select(pl.col('acronym').alias('names'), pl.col('id'), pl.col('parent_structure_id').cast(int))
        join_df = name.join(name, left_on='parent_structure_id', right_on='id')
        parent_child = join_df.select(pl.col('names'), pl.col('names_right').alias('parents'))

        return parent_child
    else:
        return df



"""
ScanBox Data Parser
====================

:author:
    Yu-Ting Wei

This module simply provides

1. result parser for scanbox acquisition system

2. view the sequence output


See all the info
-------------------

.. code-block:: python

    from neuralib.calimg.scanbox import SBXInfo

    file = ...  # scanbox .mat output file
    sbx = SBXInfo.load(file)
    sbx.print_asdict()  # print the information as dictionary



Save as Json
-------------------

.. code-block:: python

   from neuralib.calimg.scanbox import sbx_to_json

   file = ...  # scanbox .mat output file
   output_file = ...  # *.json
   sbx_to_json(file, outputfile)




Screen Shot file to tiff
-------------------------

.. code-block:: python

    from neuralib.calimg.scanbox import screenshot_to_tiff

    file = ...  # scanbox .mat screenshot output file
    output = ... # *.tiff
    screenshot_to_tiff(file, output)



SBXViewer
==================

directly view the image sequence as mmap


Use CLI
------------

See help::

    python -m neuralib.calimg.scanbox.viewer -h


Example playing the 100-200 frames::

    python -m neuralib.calimg.scanbox.viewer -D <DIR> -P <OPTIC_PLANE> -C <PMT_CHANNEL> -F 100,200


Example save 100-200 Frames as tiff::

    python -m neuralib.calimg.scanbox.viewer -D <DIR> -P <OPTIC_PLANE> -C <PMT_CHANNEL> -F 100,200 -O test.tiff



Use API call
--------------

.. code-block:: python

    from neuralib.calimg.scanbox.viewer import SBXViewer

    directory  = ...  # directory contain the .sbx and .mat output from scanbox
    sbx_viewer = SBXViewer(directory)

    # play
    sbx_viewer.play(slice(100,200), plane=0, channel=0)

    # save as tiff
    sbx_viewer.to_tiff(slice(100,200), plane=0, channel=0, output='test.tiff')


"""

from .core import *
from .viewer import *



from typing import Iterable, Literal

import attrs
import numpy as np
from matplotlib.patches import Polygon
from typing_extensions import TypeAlias, Self

__all__ = ['ObjectiveFov']

Coordinates: TypeAlias = np.ndarray
"""Array[float, 2] in XY"""


@attrs.define
class ObjectiveFov:
    """
    Class for 2P Field Of View coordinates

    Use IBL coordinates space ::

        AP (+), anterior to the Bregma. (-) posterior to the Bregma

        ML (+), right hemisphere. (-) left hemisphere

    .. seealso::  `<https://int-brain-lab.github.io/iblenv/notebooks_external/atlas_working_with_ibllib_atlas.html#Coordinate-systems>`_


    """

    region_name: str
    """FOV name"""
    am: Coordinates
    """anteromedial (default in mm). `Array[float, 2]` in XY"""
    pm: Coordinates
    """posteromedial (default in mm). `Array[float, 2]` in XY"""
    al: Coordinates
    """anterolateral (default in mm). `Array[float, 2]` in XY"""
    pl: Coordinates
    """posterolateral (default in mm). `Array[float, 2]` in XY"""

    #
    rotation_angle_ml: float = attrs.field(default=0, kw_only=True)
    """objective ml direction rotation in deg"""
    rotation_angle_ap: float = attrs.field(default=0, kw_only=True)
    """objective ap direction rotation in deg"""

    #
    unit: Literal['mm', 'um'] = attrs.field(default='mm', kw_only=True, validator=attrs.validators.in_(('mm', 'um')))

    def __iter__(self) -> Iterable[Coordinates]:
        for c in (self.am, self.pm, self.pl, self.al):  # order for polygon
            yield c

    @property
    def ap_distance(self) -> float:
        """anterior to posterior distance of the FOV"""
        a = np.max(self.am[1], self.al[1])
        p = np.max(self.pm[1], self.pl[1])
        return np.abs(p - a)

    @property
    def ml_distance(self) -> float:
        """medial to lateral distance of the FOV"""
        m = np.max(self.am[0], self.pm[0])
        l = np.max(self.al[0], self.pl[0])  # noqa: E741
        return np.abs(l - m)

    def to_um(self) -> Self:
        """unit from mm to um"""
        if self.unit == 'um':
            raise RuntimeError('unit already in um')

        factor = 1000
        return attrs.evolve(
            self,
            am=self.am * factor,
            pm=self.pm * factor,
            al=self.al * factor,
            pl=self.pl * factor,
            unit='um'
        )

    def ap_invert(self) -> Self:
        """anterior posterior coordinates invert"""
        for c in self:
            c[1] = c[1] * -1
        return attrs.evolve(self, self.am, self.pm, self.al, self.pl)

    def ml_invert(self) -> Self:
        """medial lateral coordinates invert"""
        for c in self:
            c[0] = c[0] * -1
        return attrs.evolve(self, self.am, self.pm, self.al, self.pl)

    def invert_axis(self) -> Self:
        """invert both ap and ml axes"""
        self.ap_invert()
        self.ml_invert()
        return attrs.evolve(self)

    def to_polygon(self, **kwargs) -> Polygon:
        """coordinates to ``matplotlib.patches.Polygon``"""
        x = [c[0] for c in self]
        y = [c[1] for c in self]
        return Polygon(list(zip(x, y)), closed=True, edgecolor='r', facecolor='none', **kwargs)



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

import numpy as np

from neuralib.typing import PathLike

__all__ = ['JsonEncodeHandler',
           'load_json',
           'save_json']


class JsonEncodeHandler(json.JSONEncoder):
    """Extend from the JSONEncoder class and handle the conversions in a default method

    **Usage**: Add kwarg, e.g., ``json_dump(..., cls=JsonEncodeHandler)``
    """

    def default(self, obj: Any) -> Any:
        """handle array/Path type"""
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, Path):
            return str(obj)
        else:
            return json.JSONEncoder.default(self, obj)


def load_json(filepath: PathLike, **kwargs) -> dict[str, Any]:
    """
    Load a json as dict
    :param filepath: json filepath
    :param kwargs: additional arguments to ``json.load()``
    :return:
    """
    with open(filepath, "r") as file:
        return json.load(file, **kwargs)


def save_json(filepath: PathLike, dict_obj: dict[str, Any], **kwargs) -> None:
    """
    Save dict as a json file

    :param filepath: json filepath
    :param dict_obj: dictionary object
    :param kwargs: additional arguments to ``json.dump()``
    """
    with open(filepath, "w") as outfile:
        json.dump(dict_obj, outfile, sort_keys=True, indent=4, cls=JsonEncodeHandler, **kwargs)



from __future__ import annotations

import numpy as np

from neuralib.util.deprecation import deprecated_class, deprecated_func


@deprecated_class(new='neuralib.util.segments.*', removal_version="0.4.0")
class Segment:
    """Segment container class"""

    def __init__(self,
                 data: Segment | np.ndarray,
                 copy: bool = True):

        if isinstance(data, Segment):
            if copy:
                self._data = data._data.copy()
            else:
                self._data = data._data
        else:
            self._data = check_segments(data, copy)

    def __repr__(self):
        """Return string representation of Segment object."""
        return f'Segment(" + {repr(self._data)} + ")'

    __str__ = __repr__

    def __len__(self):
        """Return the number of segments in the container."""
        return int(self._data.shape[0])

    def __getitem__(self, item: slice | int) -> Segment:
        """

        :param item: slice or index
        :return:
        """
        return Segment(self._data[item, :])

    def __iter__(self):
        """Iterate through segments in container."""
        idx = 0
        while idx < self._data.shape[0]:
            yield self._data[idx, 0], self._data[idx, 1]
            idx += 1

    @classmethod
    def from_logical(cls,
                     y: np.ndarray,
                     x: np.ndarray | None = None,
                     interpolate=False) -> Segment:
        """Construct Segment from logical vector

        :param y: 1d logical array
            Any sequence of True values that is flanked by False values is
            converted into a segment
        :param x: 1d array like, optional
            The segment indices from y will be used to index into x
        :param interpolate: bool, optional
            if true, segments of duration 0 are extent to have a minimal duration

        """
        y = np.where(y, 1, 0).ravel()

        if len(y) == 0 or np.all(y == 0):
            return cls(np.zeros((0, 2), dtype=np.int64))

        offset = 0
        if interpolate:
            offset = 0.5

        d = np.diff(np.concatenate(([0], y, [0])))
        segstart = np.nonzero(d[0:-1] == 1)[0] - offset
        segend = np.nonzero(d[1:] == -1)[0] + offset

        seg = np.vstack((segstart, segend)).T

        if x is not None:
            if interpolate:
                from scipy.interpolate import interp1d
                seg = interp1d(
                    np.arange(len(y)),
                    x,
                    kind="linear",
                    bounds_error=False,
                    fill_value=(x[0], x[-1]),
                )(seg)
            else:
                seg = x[seg]

        return cls(seg)

    @property
    def start(self) -> np.ndarray:
        """Get/Set vector of segment start values
        """
        return self._data[:, 0].copy()

    @property
    def stop(self) -> np.ndarray:
        return self._data[:, 1]


@deprecated_func(new='neuralib.util.segments.*', removal_version="0.4.0")
def check_segments(x: np.ndarray, copy=False) -> np.ndarray:
    """Convert to segment array.

    :param x: 1d array-like or (n,2) array-like
    :param copy: the output will always be a copy of the input
    :return: (n,2) array
    """
    try:
        x = np.array(x, copy=copy)
    except TypeError:
        raise ValueError("Cannot convert data to numpy array")

    if not np.issubdtype(x.dtype, np.number):
        raise ValueError("Values are not real numbers")

    # The array has to have two dimensions of shape(X,2), where X>=0.
    # As a special case, a one dimensional vector of at least length two is considered
    # a valid list of segments, e.g. when data is specified as a list [0,2,3].

    if x.shape == (0,):
        x = np.zeros([0, 2])
    elif x.ndim == 1 and len(x) > 1:
        x = np.vstack((x[0:-1], x[1:])).T
    elif x.ndim != 2 or x.shape[1] != 2:
        raise ValueError("Incorrect array size")

    # Negative duration segments are not allowed.
    if np.any(np.diff(x, axis=1) < 0):
        raise ValueError("Segment durations cannot be negative")

    return x



from typing import NamedTuple

import numpy as np
from scipy import stats
from scipy.interpolate import interp1d
from typing_extensions import Self

from neuralib.util.segments import segment_epochs, segment_duration, segment_contains
from neuralib.util.unstable import unstable

__all__ = [
    'CircularPosition',
    'interp_pos1d',
    #
    'speed_2d',
    'direction_2d',
    'interp_gap2d'
]


class CircularPosition(NamedTuple):
    """
    Position information in circular environment.

    `Dimension parameters`:

        P = Number of position points (after interpolation)

        T = Number of trial(lap)

    """

    t: np.ndarray
    """1D time array in s. `Array[float, P]`"""

    p: np.ndarray
    """1D position array in cm. `Array[float, P]`"""

    d: np.ndarray
    """1D displacement array in cm. `Array[float, P]`"""

    v: np.ndarray
    """1D velocity array in cm/s. `Array[float, P]`"""

    trial_time_index: np.ndarray
    """1D time index for every trial. `Array[int, T]`"""

    def with_time_range(self, t0: float, t1: float) -> Self:
        """With specific time range

        :param t0: time start
        :param t1: time end
        :return: A new instance of the class with updated subset of attributes (t, p, d, v) based on the time range.
        """
        tx0 = self.t >= t0
        tx1 = self.t <= t1
        tx = np.logical_and(tx0, tx1)

        ix0 = np.nonzero(tx0)[0][0]  # start index for time
        ix1 = np.nonzero(~tx1)[0][0]  # end index for time
        ix = np.logical_and(self.trial_time_index >= ix0, self.trial_time_index <= ix1)

        return self._replace(
            t=self.t[tx],
            p=self.p[tx],
            d=self.d[tx],
            v=self.v[tx],
            trial_time_index=self.trial_time_index[ix]
        )

    def with_run_mask1d(self, **kwargs) -> Self:
        """With only the running epoch. **Note that shape will be changed (stationary epoch excluded)**

        :param kwargs: Additional keyword arguments to pass to running_mask1d function.
        :return: A new instance of the class with updated subset of attributes (t, p, d, v) based on the mask generated.
        """
        from .epoch import running_mask1d
        x = running_mask1d(self.t, self.v, **kwargs)
        return self._replace(
            t=self.t[x],
            p=self.p[x],
            d=self.d[x],
            v=self.v[x],
        )

    @property
    def trial_array(self) -> np.ndarray:
        """Trial number array as same shape as *P*

        :return: `Array[int, P]`
        """
        ret = np.zeros_like(self.t, int)
        ret[self.trial_time_index] = 1
        return np.cumsum(ret)


def interp_pos1d(time: np.ndarray,
                 pos: np.ndarray,
                 *,
                 sampling_rate: float = 1000,
                 norm_max_value: float = 150,
                 remove_nan: bool = True,
                 renew_trial_value: float = -100) -> CircularPosition:
    """
    Interpolate the raw position data

    `Dimension parameters`:

        P = Number of position points (Raw)

    :param time: An array of time stamps corresponding to positional data. `Array[float, P]`.
    :param pos: An array of positional data (a.u, encoder or other hardware readout). `Array[float, P]`.
    :param sampling_rate: The rate at which data points should be sampled, defaults to 1000.
    :param norm_max_value: The value by which to normalize the positional data, defaults to 150 (cm).
    :param remove_nan: Flag indicating whether to remove NaN values from interpolated data, defaults to True.
    :param renew_trial_value: The value used to detect the overflow index which signifies trial renewals, defaults to -100.
    :return: ``CircularPosition``
    """

    overflow_idx = np.nonzero(np.diff(pos) < renew_trial_value)[0]

    # Get the position(p), displacement(d)
    if len(overflow_idx) == 0:  # not more than a trial
        p = pos / np.max(pos)
        d = pos / np.max(pos)
    else:
        p = np.zeros_like(pos, float)
        d = np.zeros_like(pos, float)
        mode = stats.mode(pos[overflow_idx], keepdims=True).mode.astype(int)

        # fix the first trial
        p[:overflow_idx[0] + 1] = pos[:overflow_idx[0] + 1] / pos[overflow_idx[0]]
        d[:overflow_idx[0] + 1] = pos[:overflow_idx[0] + 1] / pos[overflow_idx[0]]
        lap_counter = 1

        # loop through rest of trials
        for lap_start, lap_stop in zip(overflow_idx[:-1], overflow_idx[1:]):
            p[lap_start + 1:lap_stop + 1] = pos[lap_start + 1:lap_stop + 1] / pos[lap_stop]
            d[lap_start + 1:lap_stop + 1] = pos[lap_start + 1:lap_stop + 1] / pos[lap_stop] + lap_counter
            lap_counter += 1

        # fix the last lap
        p[overflow_idx[-1] + 1:] = p[overflow_idx[-1] + 1:] / mode
        d[overflow_idx[-1] + 1:] = p[overflow_idx[-1] + 1:] / mode + lap_counter

    # actual length
    p *= norm_max_value
    d *= norm_max_value

    # interpolate
    t0 = np.min(time)
    t1 = np.max(time)
    tt = np.linspace(t0, t1, int((t1 - t0) * sampling_rate))
    pp = interp1d(time, p, kind='nearest', copy=False, bounds_error=False, fill_value=np.nan)(tt)
    dd = interp1d(time, d, kind='linear', copy=False, bounds_error=False, fill_value='extrapolate')(tt)
    vv = np.diff(dd, prepend=dd[0]) * sampling_rate

    # remove nan from position_value (from interpolation boundary error)
    if remove_nan:
        x = ~np.isnan(pp)
        tt = tt[x]
        pp = pp[x]
        dd = dd[x]
        vv = vv[x]

    # time index foreach trial
    if len(overflow_idx) == 0:
        trial_time_index = np.array([0])
    else:
        trial_time_index = np.zeros_like(overflow_idx, int)
        for i, t in enumerate(time[overflow_idx]):
            # shape of t changed after interpolation(tt), find the minimal tt right after the t
            trial_time_index[i] = np.nonzero(tt >= t)[0][0]

    return CircularPosition(tt, pp, dd, vv, trial_time_index)


# =========== #
# 2D Movement #
# =========== #

def _complex_2d(xy: np.ndarray, dt: float = 1.0) -> np.ndarray:
    return np.gradient(xy[:, 0], dt) + np.gradient(xy[:, 1], dt) * 1.0j


def speed_2d(xy: np.ndarray, dt: float = 1.0) -> np.ndarray:
    """
    Compute speed 2D movement

    :param xy: 2D array coordinates in xy. `Array[float, [N, 2]]`
    :param dt: Time period between samples (for smoothing)
    :return: Speed. `Array[float, N]`
    """
    return np.abs(_complex_2d(xy, dt))


def direction_2d(xy: np.ndarray, dt: float = 1.0) -> np.ndarray:
    """
    Compute direction 2D movement

    :param xy: 2D array coordinates in xy. `Array[float, [N, 2]]`
    :param dt: Time period between samples (for smoothing)
    :return: Movement direction. `Array[float, N]`
    """
    return np.angle(_complex_2d(xy, dt), deg=True)


@unstable()
def interp_gap2d(t: np.ndarray,
                 xy: np.ndarray,
                 duration: float) -> np.ndarray:
    """
    Interpolate over gaps in position record.

    :param t: Time Vector. `Array[float, N]`
    :param xy: 2D array coordinates in xy. `Array[float, [N, 2]]`
    :param duration: Maximum duration of a gap that should be interpolated over. same unit as ``t``
    :return: Corrected x,y coordinates. `Array[float, [N, 2]]`
    """
    invalid = np.isnan(xy[:, 0])
    valid = ~invalid

    # select small gaps
    invalid_seg = segment_epochs(invalid, t)
    invalid_seg = invalid_seg[segment_duration(invalid_seg) < duration]

    if invalid_seg.shape[0] > 0:
        print(f'{len(invalid_seg)} segments invalid')
        invalid_indices = np.nonzero(segment_contains(invalid_seg, t))[0]

        f = interp1d(
            t[valid],
            xy[valid, :],
            kind="linear",
            bounds_error=False,
            axis=0,
            assume_sorted=True,
        )
        xy[invalid_indices, :] = f(t[invalid_indices])

    return xy



from __future__ import annotations

import functools
import inspect

from typing_extensions import LiteralString

from neuralib.util.func import create_fn
from . import expr

__all__ = ['as_func_expr']


def as_func_expr(f=None, *, func=expr.SqlFuncOper):
    # noinspection PyShadowingNames
    def _as_func_expr(f):
        func_name = f.__name__.upper()
        s = inspect.signature(f)
        para = []
        code = []
        locals = dict(__oper__=func, __func__=f)

        n_none = 0

        for n, p in s.parameters.items():
            if p.kind == inspect.Parameter.VAR_POSITIONAL:
                para.append('*' + n)
                n_none = -1
                break
            elif p.default is inspect.Parameter.empty:
                para.append(n)
            elif p.default is None:
                para.append((n, 'None'))
                n_none += 1
            else:
                raise RuntimeError()

            if p.annotation == 'LiteralString' or p.annotation == LiteralString:
                code.append(f'{n} = __literal__(repr({n}))')
                locals['__literal__'] = expr.SqlLiteral

        if func == expr.SqlFuncOper:
            func_call = '__func__,'
        else:
            func_call = ''

        if n_none == -1:
            args = ', '.join(para)
            code.append(f'return __oper__("{func_name}", {func_call} {args})')
        elif n_none == 0:
            args = ', '.join(para)
            code.append(f'return __oper__("{func_name}", {func_call} {args})')
        elif n_none == 1:
            assert isinstance(para[-1], tuple)
            z = para[-1][0]
            code.append(f'if {z} is None:')
            args = ', '.join(para[:-1])
            code.append(f'  return __oper__("{func_name}", {func_call} {args})')
            code.append('else:')
            args = args + ', ' + z
            code.append(f'  return __oper__("{func_name}", {func_call} {args})')
        else:
            args = ', '.join(para[:-n_none])
            for i in range(n_none, 0, -1):
                assert isinstance(para[-i], tuple)
                z = para[-i][0]
                code.append(f'if {z} is None:')
                code.append(f'  return __oper__("{func_name}", {func_call} {args})')
                args = args + ', ' + z
            code.append(f'return __oper__("{func_name}", {func_call} {args})')

        ret = create_fn(f.__name__, para, '\n'.join(code), locals=locals)
        return functools.wraps(f)(ret)

    if f is None:
        return _as_func_expr
    else:
        return _as_func_expr(f)



import abc
from pathlib import Path
from typing import Iterable

import cv2
import numpy as np
from csbdeep.utils import normalize

from neuralib.argp import AbstractParser, argument

__all__ = ['AbstractSegmentationOption']


class AbstractSegmentationOption(AbstractParser, metaclass=abc.ABCMeta):
    DESCRIPTION = 'ABC for cellular segmentation'

    file: Path = argument('-F', '--file', help='image file path')
    directory: Path = argument('-D', '--dir', help='images directory for batch processing')
    directory_suffix: str = argument('--suffix', default='.tif',
                                     choices=['.tif', '.tiff', '.png'],
                                     help='suffix in batch mode')
    save_ij_roi: bool = argument('--ij-roi', '--roi', help='if save also the imageJ/Fiji compatible .roi file')
    force_re_eval: bool = argument('--force-eval', '--re', help='force re-evaluate the result')

    model: str = argument('-M', '--model', metavar='MODEL', help='which pretrained model')
    no_normalize: bool = argument('--no-norm', help='NOT DO Percentile-based image normalization for eval')
    napari_view: bool = argument('--napari', help='view in napari')

    def post_parsing(self):
        """check args is valid"""
        if not self.file_mode and not self.batch_mode:
            raise ValueError('Either use "-F in file" or "-D in directory" for processing')

        if self.file and self.directory:
            raise ValueError('Only "-F in file" or "-D in directory" should be specified, not both')

        if self.directory and self.napari_view:
            raise ValueError('napari view only used in single file mode')

    @property
    def file_mode(self) -> bool:
        """Flag file mode"""
        return self.file is not None

    @property
    def batch_mode(self) -> bool:
        """Flag batch mode"""
        return self.directory is not None

    @staticmethod
    def _as_grayscale(file: str) -> np.ndarray:
        img = cv2.imread(file)
        if img.ndim == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        return img

    def raw_image(self) -> np.ndarray:
        """Load image from file and convert to grayscale

        :return: `Array[float, [H, W]]`
        """
        if not self.file.is_file():
            raise ValueError('not a file')

        return self._as_grayscale(str(self.file))

    def normalize_image(self) -> np.ndarray:
        """
        Normalize the image

        :return: `Array[float, [H, W]]`
        """
        img = self.raw_image()

        return normalize(img, clip=True)

    def foreach_raw_image(self) -> Iterable[tuple[Path, np.ndarray]]:
        """
        Load image from a directory and convert to grayscale

        :return: Tuple of filepath and image `Array[float, [H, W]]`
        """
        for file in self.directory.glob(f'*{self.directory_suffix}'):
            img = self._as_grayscale(str(file))
            yield file, img

    def foreach_normalize_image(self) -> Iterable[tuple[Path, np.ndarray]]:
        """
        Normalize the image in batch mode

        :return: Tuple of filepath and image `Array[float, [H, W]]`
        """
        for name, raw in self.foreach_raw_image():
            yield name, normalize(raw, clip=True)

    @abc.abstractmethod
    def seg_output(self, filepath: Path) -> Path:
        """
        Get segmented output save path

        :param filepath: filepath for image
        :return: segmented output save path
        """
        pass

    def ij_roi_output(self, filepath: Path) -> Path:
        """
        Get imageJ/Fiji ``.roi`` output save path

        :param filepath: filepath for image
        :return: ij roi output save path
        """
        return filepath.with_suffix('.roi')

    @abc.abstractmethod
    def eval(self) -> None:
        """eval the model in single file or batch files, and save the results"""
        pass

    @abc.abstractmethod
    def launch_napari(self, **kwargs):
        """napari viewer"""
        pass



from __future__ import annotations

from typing import Literal, NamedTuple

import numpy as np
from scipy.ndimage import gaussian_filter, gaussian_filter1d, minimum_filter1d, maximum_filter1d

from .core import Suite2PResult, SIGNAL_TYPE

__all__ = [
    'get_neuron_signal',
    'normalize_signal',
    'normalize_signal_factor',
    #
    'DFFSignal',
    'dff_signal',
    'calc_signal_baseline',
    #
    'sync_s2p_rigevent',
]

BASELINE_METHOD = Literal['maximin', 'constant', 'constant_prctile']


def get_neuron_signal(s2p: Suite2PResult,
                      n: int | np.ndarray | list[int] | None = None,
                      *,
                      signal_type: SIGNAL_TYPE = 'df_f',
                      normalize: bool = True,
                      dff: bool = True,
                      correct_neuropil: bool = True,
                      method: BASELINE_METHOD = 'maximin') -> tuple[np.ndarray, np.ndarray]:
    """
    Select neuronal signals for analysis. For single cell (F,) OR multiple cells (N, F)

    :param s2p: suite 2p result
    :param n: neuron index (`int`) or index arraylike (`Array[int, N]`). If None, then use all neurons
    :param signal_type: signal type. :data:`~neuralib.calimg.suite2p.core.SIGNAL_TYPE` {'df_f', 'spks'}
    :param normalize: 01 normalization for each neuron
    :param dff: normalize to the baseline fluorescence changed (dF/F)
    :param correct_neuropil: do the neuropil correction
    :param method: baseline calculation method {'maximin', 'constant', 'constant_prctile'}
    :return: tuple with (signal, baseline signal). `Array[float, F|[N,F]]`
    """
    if n is None:
        n = list(range(s2p.n_neurons))

    f = s2p.f_raw[n]
    fneu = s2p.f_neu[n]

    if signal_type == 'df_f':

        if dff:
            s1 = dff_signal(f, fneu, s2p, correct_neuropil, method).dff
            s2 = np.full_like(s1, 0, dtype=int)
            if normalize:
                o, f = normalize_signal_factor(s1)
                s1 = (s1 - o) / f

        else:
            fcorr = f - 0.7 * fneu
            sig_baseline = s2p.signal_baseline
            window = int(s2p.window_baseline * s2p.fs)
            s1 = fcorr
            # baseline not a constant value, due to calcium decay or motion drift
            s2 = _maximin_filter(fcorr, sig_baseline, window)
            if normalize:
                o, f = normalize_signal_factor(s1)
                s1 = (s1 - o) / f
                ob, fb = normalize_signal_factor(s2)
                s2 = (s2 - ob) / fb

    elif signal_type == 'spks':
        s1 = s2p.spks[n]
        s2 = np.full_like(s1, 0, dtype=int)

        if normalize:
            s1 = normalize_signal(s1)

    else:
        raise ValueError('specify signal type either in df_f or spks')

    return s1, s2


def normalize_signal(s: np.ndarray, axis: int = -1) -> np.ndarray:
    """
    do 0 1 normalization

    :param s: signal
    :param axis: axis
    :return: normalized signal
    """
    o, f = normalize_signal_factor(s, axis)
    return (s - o) / f


def normalize_signal_factor(s: np.ndarray, axis: int = -1) -> tuple[np.ndarray, np.ndarray]:
    return np.min(s, axis=axis, keepdims=True), \
        np.max(s, axis=axis, keepdims=True) - np.min(s, axis=axis, keepdims=True)


# ============ #
# DF/F Compute #
# ============ #

class DFFSignal(NamedTuple):
    """Container for dF/F signal processing.

    `Dimension parameters`:

        N = number of neurons

        F = number of frame

    For single cell (F,) OR multiple cells (N, F)

    """

    s2p: Suite2PResult
    """Suite2PResult"""
    f: np.ndarray
    """Fluorescence intensity. `Array[float, F | [N,F]]`"""
    fneu: np.ndarray
    """Neuropil intensity. `Array[float, F | [N,F]]`"""
    fcorr: np.ndarray
    """Fluorescence corrected by fneu that used for dff calculation. `Array[float, F | [N,F]]`"""
    f0: np.ndarray
    """Background Fluorescence. `Array[float, F | [N,F]]`"""
    dff: np.ndarray
    """dff after f0 normalization. `Array[float, F | [N,F]]`"""

    @property
    def dff_baseline(self) -> np.ndarray:
        """baseline of dff, supposed to be 0"""
        return np.full_like(self.dff, 0)

    @property
    def baseline_fluctuation(self) -> np.ndarray:
        """get the fluctuation of the fneu signal. `Array[float, F | [N,F]]`

        Perhaps not fully corrected with physiological reason.
        **used to get the baseline std (i.e., trial reliability metric)**
        """
        fneu_corr = 0.3 * self.fneu
        fneu_bas = calc_signal_baseline(fneu_corr, self.s2p, method='maximin')
        return 100 * ((fneu_corr - fneu_bas) / fneu_bas)


def dff_signal(f: np.ndarray,
               fneu: np.ndarray,
               s2p: Suite2PResult,
               correct_neuropil: bool = True,
               method: BASELINE_METHOD = 'maximin') -> DFFSignal:
    """
    df_f signal normalization container

    :param f: Fluorescence intensity. `Array[float, F | [N,F]]`
    :param fneu: Neuropil intensity. `Array[float, F | [N,F]]`
    :param s2p: ``Suite2PResult``
    :param correct_neuropil: whether do the subtraction using neuropil
    :param method: {'maximin', 'constant', 'constant_prctile'}

    :return: DFFSignal
    """

    fcorr = f - 0.7 * fneu if correct_neuropil else f

    f0 = calc_signal_baseline(fcorr, s2p, method)
    dff = 100 * ((fcorr - f0) / f0)
    return DFFSignal(s2p, f, fneu, fcorr, f0, dff)


def calc_signal_baseline(signal: np.ndarray,
                         s2p: Suite2PResult,
                         method: BASELINE_METHOD = 'maximin') -> np.ndarray:
    """
    f0 calculation

    .. seealso:: source code from suite2p: ``suite2p.extraction.dcnv.preprocess``

    :param signal: signal activity. i.e., fcorr
    :param s2p: :class:`~neuralib.calimg.suite2p.core.Suite2PResult`
    :param method: BASELINE_METHOD
    :return:
        f0 baseline for the df/f calculation
    """

    if method == 'maximin':
        sig_baseline = s2p.signal_baseline
        window = int(s2p.window_baseline * s2p.fs)
        f0 = _maximin_filter(signal, sig_baseline, window)

    elif method == 'constant':
        signal = gaussian_filter(signal, [0., s2p.signal_baseline])
        f0 = np.min(signal, axis=-1)

    elif method == 'constant_prctile':
        sig_prc = np.percentile(signal, s2p.prctile_baseline, axis=-1)
        f0 = np.expand_dims(sig_prc, axis=-1)

    else:
        raise TypeError(f'unknown method{method}')

    return f0


def _maximin_filter(signal: np.ndarray, kernel: float, size: int) -> np.ndarray:
    """
    max/min method to calculate the signal baseline.
    **Note that this method is sensitive to the kernel size

    :param signal:
    :param kernel: standard deviation for Gaussian kernel
    :param size: length along which to calculate 1D minimum
    :return:
    """
    signal = gaussian_filter1d(signal, kernel, axis=-1)
    signal = minimum_filter1d(signal, size, axis=-1)
    baseline = maximum_filter1d(signal, size, axis=-1)

    return baseline


# ================= #
# Sync Pulse Number #
# ================= #

def sync_s2p_rigevent(image_time: np.ndarray,
                      s2p: Suite2PResult,
                      plane: int | None = 0) -> np.ndarray:
    r"""

    Do the alignment to make the shape the same in hardware pulses `(P,)` & suite2p output files `(F,)`

    \* ASSUME both recordings are sync

    \* ASSUME sequential scanning pattern. i.e., 0, 1, 2, 3 ETL scanning order

    :param image_time: imaging time hardware pulses. `Array[float, P]`
    :param s2p: :class:`~neuralib.calimg.suite2p.core.Suite2PResult`
    :param plane: number of optical plane
    :return: image_time: with same len as s2p results `Array[float, F]`
    """
    n_plane = s2p.n_plane
    image_fr = s2p.fs

    if plane is not None:
        image_time = image_time[plane::n_plane]
    else:
        # TODO used npy in combined file
        raise NotImplementedError('')

    n_image = len(image_time)
    n_frame = s2p.n_frame
    n_diff = abs(n_image - n_frame)

    if n_diff >= n_frame / 100:
        raise RuntimeError('too large difference between riglog imaging header and calcium data')

    if n_image > n_frame:
        # drop latest, mostly this case
        image_time = image_time[:n_frame]
    elif n_image < n_frame:
        # padding at the tail
        image_time = np.hstack((image_time, [image_time[-1] + (i + 1) / image_fr for i in range(n_diff)]))

    return image_time



from __future__ import annotations

import attrs

from neuralib.util.deprecation import deprecated_func

__all__ = ['CliArgs']


@deprecated_func(remarks='over engineering', removal_version="0.4.0")
@attrs.define
class CliArgs:
    flag: str | None = attrs.field(default=None)
    args: str | None = attrs.field(default=None)

    # noinspection PyUnresolvedReferences
    @flag.validator
    def _check_flag(self, attribute, value: str | None) -> None:
        if value is None:
            return

        if not isinstance(value, str):
            raise TypeError(f'flag not a str: "{value}"')

        valid1 = value.startswith('-') and len(value) > 1
        valid2 = value.startswith('--') and len(value) > 2
        if not (valid1 or valid2):
            raise ValueError(f'{self} {attribute.name} not valid')

    # noinspection PyUnresolvedReferences
    @args.validator
    def _check_args(self, attribute, value: str) -> None:
        if value is None:
            return

        if not isinstance(value, str):
            self.args = str(value)

    def as_command(self) -> list[str]:

        # position arg
        if self.flag is None and self.args is not None:
            return [self.args]

        # flag
        elif self.flag is not None and self.args is None:
            return [self.flag]

        # flag + arg
        elif self.flag is not None and self.args is not None:
            return [self.flag, self.args]
        else:
            raise RuntimeError('')

    @classmethod
    def concat_command(cls, args: list[CliArgs]) -> list[str]:
        """concat multiple ``CliArgs`` as list of subprocess command line"""
        ret = []
        for arg in args:
            ret.extend(arg.as_command())

        return ret



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import Literal, Union, Callable

import pandas as pd
import polars as pl
from colorama import Fore, Style

from neuralib.typing import DataFrame

__all__ = ['fprint',
           'printdf',
           'print_load',
           'print_save',
           'publish_annotation']

_PREV_LOAD_FILE = None
_PREV_SAVE_FILE = None


def fprint(*msgs,
           vtype: Literal['info', 'io', 'warning', 'error', 'pass', 'debug'] | str = 'info',
           timestamp: bool = True,
           **kwarg) -> None:
    """
    Formatting print with different colors based on verbose type

    :param msgs: print message
    :param vtype: verbose type
    :param timestamp: If add time info
    """
    prefix = f'[{vtype.upper()}]'

    if vtype == 'error':
        color = 'RED'
    elif vtype == 'warning':
        color = 'YELLOW'
    elif vtype == 'io':
        color = 'MAGENTA'
    elif vtype == 'info':
        color = 'CYAN'
    elif vtype == 'pass':
        color = 'GREEN'
    else:
        color = 'WHITE'

    try:
        fg_color = getattr(Fore, color)
    except AttributeError:
        fg_color = Fore.WHITE

    msg = fg_color + prefix
    if timestamp:
        msg += f"[{datetime.today().strftime('%y-%m-%d %H:%M:%S')}] - "

    try:
        out = ''.join(msgs) + "\n"
    except TypeError:
        out = f'{msgs}'

    msg += out
    msg += Style.RESET_ALL
    print(msg, **kwarg)


def printdf(df: DataFrame,
            nrows: int | None = None,
            ncols: int | None = None,
            tbl_width_chars: int = 500,
            do_print: bool = True) -> str:
    """
    print dataframe with given row numbers (polars)
    if isinstance pandas dataframe, print all.

    :param df: polars or pandas dataframe
    :param nrows: number of rows (applicable in polars case)
    :param ncols: number of columns
    :param tbl_width_chars: table width for showing
    :param do_print: do print otherwise, only return the str
    :return:
    """

    if isinstance(df, pl.DataFrame):
        with pl.Config(tbl_width_chars=tbl_width_chars) as cfg:
            rows = df.shape[0] if nrows is None else nrows
            cols = df.shape[1] if ncols is None else ncols
            cfg.set_tbl_rows(rows)
            cfg.set_tbl_cols(cols)

            if do_print:
                print(df)

            return df.__repr__()

    elif isinstance(df, pd.DataFrame):
        ret = df.to_markdown()
        print(ret)
        return ret

    else:
        raise TypeError('')


def print_load(file: Union[str, Path], verb='LOAD') -> Path:
    """print message for loading file.

    If *file* is not existed, '!' will add after *verb*.
    File load message doesn't print twice and more if it shows continuously, so it is okay
    that wrap a loading method and call :func:`print_load` on target file.

    use flag "load".

    :param file: file path
    :param verb: default 'LOAD'
    :return: *file*
    """
    global _PREV_LOAD_FILE

    file = Path(file)

    if _PREV_LOAD_FILE != file:
        if len(verb) < 4:
            verb += ' ' * (4 - len(verb))

        not_exist = '!' if not file.exists() else ' '

        from tqdm import tqdm
        with tqdm.external_write_mode():
            fprint(f'{verb}{not_exist}->{file}', vtype='io')

    _PREV_LOAD_FILE = file
    return file


def print_save(file: Union[str, Path], verb='SAVE') -> Path:
    """print message for saving file.

    If *file* is not existed, '+' will add after *verb*.
    File save message doesn't print twice and more if it shows continuously, so it is okay
    that wrap a saving method and call :func:`print_load` on target file.

    use flag "save".

    :param file: file path
    :param verb: default 'SAVE'
    :return: *file*
    """
    global _PREV_SAVE_FILE

    file = Path(file)

    if _PREV_SAVE_FILE != file:
        if len(verb) < 4:
            verb += ' ' * (4 - len(verb))

        not_exist = '+' if not file.exists() else ' '

        from tqdm import tqdm
        with tqdm.external_write_mode():
            fprint(f'{verb}{not_exist}->{file}', vtype='io')

    _PREV_SAVE_FILE = file
    return file


def publish_annotation(level: Literal['main', 'sup', 'appendix', 'test'],
                       *,
                       project: str | list[str] | None = None,
                       figure: str | list[str] | None = None,
                       caption: str | None = None,
                       as_doc: bool = False,
                       as_attributes: bool = True):
    """
    Annotation for knowing the class/function using scenario in paper publication

    :param level: {'main', 'sup', 'appendix', 'test'}
    :param project: Project name or list of project name
    :param figure: Figure number or list of figure name
    :param caption: Other caption
    :param as_doc: As documents, be able to parser the `.. note::` block by ``Sphinx``
    :param as_attributes: If set info as attributes
    :return:
    """
    valid_levels = ('main', 'sup', 'appendix', 'test')
    if level not in valid_levels:
        raise ValueError(f'must be one of the {valid_levels}')

    def decorator(target: type | Callable):

        doc = '' if target.__doc__ is None else target.__doc__

        if as_doc:
            doc += '\n\n' + '\n\n'.join([
                f'.. note:: ',
                f'\t**Publish Annotation**',
                f'\tProject: {project}',
                f'\tFigure: {figure}',
                f'\tLevel: {level} ',
                f'\tCaption: {caption}',
            ])
            target.__doc__ = doc

        if as_attributes:
            attrs = ['__publish_level__', '__publish_project__', '__publish_figure__', '__publish_caption__']

            for attr in attrs:
                if hasattr(target, attr):
                    raise AttributeError(f"Class {target.__name__} already has an attribute named '{attr}'.")

            target.__publish_level__ = level
            target.__publish_project__ = project
            target.__publish_figure__ = figure
            target.__publish_caption__ = caption

        return target

    return decorator



from pathlib import Path
from typing import Literal, final

import attrs
import napari
import numpy as np
from numpy.lib.npyio import NpzFile
from stardist.models import StarDist2D
from typing_extensions import Self

from neuralib.argp import argument, as_argument
from neuralib.segmentation.base import AbstractSegmentationOption
from neuralib.typing import PathLike
from neuralib.util.color_logging import setup_clogger, LOGGING_IO_LEVEL

STARDIST_MODEL = Literal['2D_versatile_fluo', '2D_versatile_he', '2D_paper_dsb2018', '2D_demo']
Logger = setup_clogger(caller_name=Path(__file__).name)

__all__ = ['StarDistResult',
           'StarDist2DOptions']


@final
@attrs.define
class StarDistResult:
    """
    Stardist results

    `Dimension parameters`:

        N = Number of segmented cell

        E = Number of polygons edge

        W = Image width

        H = Image height

        P = Number of image pixel with label

    """
    filename: str
    """Source image file name"""

    labels: np.ndarray
    """Image with label. `Array[float, [H, W]]`"""  # TODO Is it float?

    cords: np.ndarray
    """Coordinates. `Array[float, [N, 2, E]]`"""

    prob: np.ndarray
    """Detected probability. `Array[float, N]`"""

    points: np.ndarray = attrs.field(init=False)
    """Coordinates to points by simple XY average. `Array[float, [N, 2]]`"""

    def __attrs_post_init__(self):
        p = np.zeros((self.cords.shape[0], 2))
        for i, c in enumerate(self.cords):
            x_mean = np.mean(c[0, :])
            y_mean = np.mean(c[1, :])
            p[i, 0] = x_mean
            p[i, 1] = y_mean
        self.points = p

    @classmethod
    def load(cls, file: PathLike) -> Self:
        dat = np.load(file)
        Logger.log(LOGGING_IO_LEVEL, f'Load stardist results from {file}')
        return cls(
            filename=dat['filename'],
            labels=cls._reconstruct_labels_from_index_value(dat),
            cords=dat['cords'],
            prob=dat['prob'],
        )

    @classmethod
    def _reconstruct_labels_from_index_value(cls, dat: NpzFile) -> np.ndarray:
        h, w = dat['shape']
        index = dat['index']
        value = dat['value']

        image = np.full((h, w), np.nan)
        for i, (hi, wi) in enumerate(index):
            image[hi, wi] = value[i]

        return image

    def savez(self, output_file: PathLike) -> None:
        """
        Save ``filename``, ``cord``, ``prob``, ``point``, ``shape``, ``index``, ``index``, ``value`` as a npz file.

        shape: `Array[int, 2] in H,W`

        index: index with labels. `Array[int, [P, 2]]`

        value: label value in its index `Array[int, P]`

        :param output_file: output
        """
        if Path(output_file).suffix != '.npz':
            raise ValueError('output file suffix need to be a .npz')

        shape = np.array(self.labels.shape)
        index, value = self.get_index_value()

        np.savez(output_file,
                 filename=self.filename,
                 cords=self.cords,
                 prob=self.prob,
                 points=self.points,
                 shape=shape,
                 index=index,
                 value=value)

        Logger.log(LOGGING_IO_LEVEL, f'save stardist results to {output_file}')

    def save_roi(self, output_file: PathLike) -> None:
        """Save as imageJ ``.roi`` file"""
        from roifile import ImagejRoi, ROI_TYPE, ROI_OPTIONS

        points = np.fliplr(self.points)  # XY rotate in .roi format
        roi = ImagejRoi(
            roitype=ROI_TYPE.POINT,
            options=ROI_OPTIONS.PROMPT_BEFORE_DELETING | ROI_OPTIONS.SUB_PIXEL_RESOLUTION,
            n_coordinates=self.points.shape[0],
            integer_coordinates=points,
            subpixel_coordinates=points
        )

        roi.tofile(output_file)

    def get_index_value(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Get labelled pixel index and its value.

        :return: `Array[int, [P, 2]]` and `Array[float, P]`
        """
        labels = self.labels
        h, w = labels.shape
        index = []
        value = []
        for i in range(h):
            for j in range(w):
                if not np.isnan(labels[i, j]):
                    index.append([i, j])
                    value.append(labels[i, j])

        return np.array(index).astype(np.int_), np.array(value).astype(np.int_)

    def with_probability(self, threshold: float) -> Self:
        m = self.prob >= threshold
        return attrs.evolve(self, cords=self.cords[m], prob=self.prob[m])


class StarDist2DOptions(AbstractSegmentationOption):
    DESCRIPTION = 'Run the Stardist model for segmentation'

    model: STARDIST_MODEL = as_argument(AbstractSegmentationOption.model).with_options(
        default='2D_versatile_fluo',
        help='stardist pretrained model'
    )

    prob_thresh: float = argument(
        '--prob',
        default=0.5,
        help='Consider only object candidates from pixels with predicted object probability above this threshold. '
             'Seealso: stardist.models.base._predict_instances_generator: prob_thresh'
    )

    def run(self):
        if self.napari_view:
            self.launch_napari()
        else:
            self.eval()

    def seg_output(self, filepath: Path) -> Path:
        """
        Get output save path

        :param filepath: filepath for image
        :return: output save path
        """
        return filepath.with_suffix('.npz')

    def eval(self, **kwargs) -> None:
        model = StarDist2D.from_pretrained(self.model)

        if self.file_mode:
            image = self.raw_image() if self.no_normalize else self.normalize_image()
            self._eval(self.file, image, model)

        elif self.batch_mode:
            iter_image = self.foreach_raw_image() if self.no_normalize else self.foreach_normalize_image()
            for name, image in iter_image:
                self._eval(name, image, model)

    def _eval(self, filepath: Path, image: np.ndarray, model: StarDist2D, **kwargs):
        labels, detail = model.predict_instances(image, prob_thresh=self.prob_thresh, **kwargs)

        labels = labels.astype(np.float_)
        labels[labels == 0] = np.nan

        res = StarDistResult(
            filepath.name,
            labels,
            detail['coord'],
            detail['prob']
        ).with_probability(self.prob_thresh)

        res.savez(self.seg_output(filepath))

        if self.save_ij_roi:
            res.save_roi(self.ij_roi_output(filepath))

    # noinspection PyTypeChecker
    def launch_napari(self, with_widget: bool = False):
        """
        Launch napari viewer for stardist results

        :param with_widget: If True, launch also with the starDist widget (required package ``stardist-napari``)
        """
        file = self.seg_output(self.file)
        if not file.exists() or self.force_re_eval:
            self.eval()

        res = StarDistResult.load(file)

        viewer = napari.Viewer()
        viewer.add_image(self.raw_image(), name='raw')
        if not self.no_normalize:
            viewer.add_image(self.normalize_image(), name='normalized')
        viewer.add_image(res.labels, name='labels', colormap='twilight_shifted', opacity=0.5)
        viewer.add_points(res.points, face_color='red')

        if with_widget:
            viewer.window.add_plugin_dock_widget("stardist-napari", "StarDist")

        Logger.info('Launch napari!')
        napari.run()


if __name__ == '__main__':
    StarDist2DOptions().main()



from __future__ import annotations

from typing import Any

from .alias import ArrayLike

__all__ = [
    'is_iterable',
    'is_namedtuple',
    'flatten_arraylike',
    'array2str'
]


def is_iterable(val: Any) -> bool:
    """Check if a value is iterable and not a string.

    :param val: The value to be checked.
    :return: True if the value is iterable and not a string, False otherwise
    """
    from collections.abc import Iterable
    return isinstance(val, Iterable) and not isinstance(val, str)


# noinspection PyProtectedMember
def is_namedtuple(obj: Any) -> bool:
    if isinstance(obj, tuple) and hasattr(obj, '_fields'):
        if isinstance(obj._fields, tuple):
            return True

    return False


def flatten_arraylike(ls: list[ArrayLike]) -> list[Any]:
    """
    flatten the list

    :param ls:
    :return: flatten list
    """
    from itertools import chain
    return list(chain(*ls))


def array2str(x: ArrayLike, sep=' ') -> str:
    """Convert an array-like object to a string with a specified separator.

    :param x: 1D array-like object.
    :param sep: Separator used to join elements. Default is a space.
    :return: A string representation of the array with elements separated by `sep`.
    """
    return sep.join(map(str, x))



import subprocess
import sys
from pathlib import Path
from typing import overload, IO, Any, Union

from .cli import Database
from .table import table_class, table_name, Table

__all__ = ['generate_dot']


@overload
def generate_dot(db: Database, *,
                 graph: dict[str, Any] = None,
                 node: dict[str, Any] = None,
                 edge: dict[str, Any] = None) -> str:
    pass


@overload
def generate_dot(db: Database, file: Union[str, Path], *,
                 graph: dict[str, Any] = None,
                 node: dict[str, Any] = None,
                 edge: dict[str, Any] = None) -> None:
    pass


@overload
def generate_dot(db: Database, file: IO, *,
                 graph: dict[str, Any] = None,
                 node: dict[str, Any] = None,
                 edge: dict[str, Any] = None) -> None:
    pass


def generate_dot(db: Database, file=None, *,
                 graph: dict[str, Any] = None,
                 node: dict[str, Any] = None,
                 edge: dict[str, Any] = None):
    """

    :param db:
    :param file: io, or an output file path which support '.puml' file and '.png' file (required ``graphivz``).
    :param graph: graph attributes
    :param node: node attributes
    :param edge: edge attributes
    :return:
    :raise FileNotFoundError: Check if graphivz already installed and add to system PATH
    """
    if file is None:
        from io import StringIO
        buf = StringIO()
        _generate_dot(db, buf, graph=graph, node=node, edge=edge)
        return buf.getvalue()
    elif isinstance(file, (str, Path)):
        file = Path(file)
        if file.suffix in ('.dot',):
            with file.open() as out:
                _generate_dot(db, out, graph=graph, node=node, edge=edge)
        elif file.suffix in ('.png', '.svg', '.ps', '.pdf', '.jpg', '.gif', '.json'):
            _generate_dot_png(generate_dot(db, graph=graph, node=node, edge=edge), file)
        else:
            raise RuntimeError(f'unsupported filetype : {file.suffix}')
    else:
        _generate_dot(db, file)


def _generate_dot_png(dot: str, file: Path) -> int:
    p = subprocess.Popen(['dot', '-T', file.suffix[1:], '-o', str(file)], stdin=subprocess.PIPE)
    p.communicate(dot.encode())
    return p.wait()


def _generate_dot(db: Database, file=sys.stdout, *,
                  graph: dict[str, Any] = None,
                  node: dict[str, Any] = None,
                  edge: dict[str, Any] = None):
    name = db.database_file
    if name is None:
        name = ':memory:'
    print('digraph', f'"{name}"', '{', file=file)

    _graph = dict()
    if graph is not None:
        _graph.update(graph)

    if len(_graph):
        print('graph [', file=file)
        for k, v in _graph.items():
            print(k, '=', v, file=file)
        print('];', file=file)

    _node = dict(shape='record', rankdir='LR')
    if node is not None:
        _node.update(node)

    if len(_node):
        print('node [', file=file)
        for k, v in _node.items():
            print(k, '=', v, file=file)
        print('];', file=file)

    _edge = dict()
    if edge is not None:
        _edge.update(edge)

    if len(_edge):
        print('edge [', file=file)
        for k, v in _edge.items():
            print(k, '=', v, file=file)
        print('];', file=file)

    _generate_dot_table(db, file)
    print('}', file=file)


def _generate_dot_table(db: Database, file=sys.stdout):
    for table in db.database_tables:
        table: Table = table_class(table)
        print(f'"{table.table_name}"', '[', file=file)

        ff = []

        for field in table.table_fields:
            at = [f'<{field.name}>']
            if field.is_primary:
                at.append('#')
            elif field.is_unique:
                at.append('!')
            at.append(field.name)
            at.append(':')
            at.append(field.sql_type.__name__)
            if not field.not_null:
                at.append('?')
            ff.append(''.join(at))

        print('label=', '"', '<0>', table.table_name, '|{', '|'.join(ff), '}"', sep='', file=file)

        print('];', file=file)

    for table in db.database_tables:
        table: Table = table_class(table)
        for foreign in table.table_foreign_fields:
            if foreign.fields == foreign.foreign_fields:
                print(table_name(foreign.table), ':0', '->', table_name(foreign.foreign_table), ':0', file=file)

            elif len(foreign.fields) == 1:
                print(table_name(foreign.table), ':', foreign.fields[0], '->', table_name(foreign.foreign_table), ':', foreign.foreign_fields[0], file=file)

            else:
                k1 = ', '.join(foreign.fields)
                k2 = ', '.join(foreign.foreign_fields)

                print(table_name(foreign.table), ':0', '->', table_name(foreign.foreign_table), ':0', '[', file=file)
                print('label=', f'"({k1}) to ({k2})"', file=file)
                print('];', file=file)



from typing import NamedTuple

import numpy as np
from sklearn.decomposition import TruncatedSVD

__all__ = ['SequenceSingularVector',
           'compute_singular_vector']


class SequenceSingularVector(NamedTuple):
    """A NamedTuple that represents a singular value decomposition result from image sequences."""

    singular_value: np.ndarray
    """The singular values of the decomposition"""
    right_vector: np.ndarray
    """Right singular vectors of the decomposition"""
    left_vector: np.ndarray
    """Left singular vectors of the decomposition"""


def compute_singular_vector(sequences: np.ndarray,
                            n_components: int = 128) -> SequenceSingularVector:
    """
    :param sequences: A numpy array representing a collection of image frames in a sequence.
                      It has a shape of (n_frames, width, height) where 'n_frames' is the number
                      of frames, and 'width' and 'height' are the dimensions of each frame.
    :param n_components: An integer representing the number of components for Truncated SVD.
                         The default value is 128.
    :return: A SequenceSingularVector object containing the singular values, the transformed
             components, and the left singular vectors.
    """
    n_frames, width, height = sequences.shape
    seq = sequences.reshape(n_frames, height * width).T
    seq = seq - np.mean(seq, axis=1, keepdims=True)

    svd = TruncatedSVD(n_components=n_components)
    Vsv = svd.fit_transform(seq.T)
    U = seq @ (Vsv / (Vsv ** 2).sum(axis=0) ** 0.5)
    U /= (U ** 2).sum(axis=0) ** 0.5

    return SequenceSingularVector(svd.singular_values_, Vsv, U)



"""

Provide CLI for cellular segmentation and visualization

Design mainly for quick visualization of results and batch processing (i.e., whole-brain image), and save as imageJ/Fiji ``.roi``.


**Supported packages**

    1. `Cellpose <https://github.com/MouseLand/cellpose>`_

    - See usage :mod:`neuralib.segmentation.cellpose`

    2. `Stardist <https://github.com/stardist/stardist>`_

    - See usage in :mod:`neuralib.segmentation.stardist`

    If only small amount of images, **the native GUI are recommended**.


"""



import subprocess

from neuralib.segmentation.cellpose.core import AbstractCellPoseOption
from neuralib.util.gpu import check_mps_available, check_nvidia_cuda_available
from neuralib.util.verbose import fprint

__all__ = ['CellPoseSubprocOption']


class CellPoseSubprocOption(AbstractCellPoseOption):
    DESCRIPTION = 'Run cellpose as a CLI subprocess. Detail refer to ``Cellpose.cli``'

    def run(self):
        self.eval()

    def eval(self) -> int:
        cmds = self.build_cli()
        code = subprocess.check_call(cmds)
        return code

    def build_cli(self) -> list[str]:
        ret = ['python', '-m', 'cellpose', '--verbose']

        #
        if self.file_mode:
            ret.extend(['--image_path', str(self.file)])
        elif self.batch_mode:
            ret.extend(['--dir', str(self.directory)])

        #
        ret.extend(['--chan', str(self.chan_seg)])
        ret.extend(['--chan2', str(self.chan_nuclear)])

        #
        ret.extend(['--pretrained_model', self.model])

        #
        ret.extend(['--diameter', str(self.diameter)])

        #
        if self.no_normalize:
            ret.extend('--no_norm')

        #
        if check_nvidia_cuda_available(backend='torch') or check_mps_available(backend='torch'):
            ret.extend('--use_gpu')

            if check_mps_available():  # use mps in macOS
                ret.extend('--gpu_device', 'mps')

        fprint(ret)

        return ret


if __name__ == '__main__':
    CellPoseSubprocOption().main()



from __future__ import annotations

import pickle
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Any, ContextManager

import gdown
import numpy as np
import polars as pl
from PIL import Image

from neuralib.calimg.scanbox import SBXInfo
from neuralib.calimg.suite2p import Suite2PResult
from neuralib.io import NEUROLIB_CACHE_DIRECTORY
from neuralib.tracking.deeplabcut.core import DeepLabCutResult, load_dlc_result
from neuralib.typing import PathLike

__all__ = [
    'google_drive_file',
    'google_drive_folder',
    #
    'load_example_scanbox',
    'load_example_suite2p',
    #
    'load_example_rois',
    'load_example_rois_image',
    #
    'load_ephys_meta',
    'load_ephys_data',
    'load_npx2_reconstruction',
    #
    'load_example_rastermap_2p',
    'load_example_rastermap_wfield',
    #
    'load_example_dlc_h5',
    'load_example_dlc_csv'
]


@contextmanager
def google_drive_file(file_id: str,
                      *,
                      quiet: bool = True,
                      output_dir: PathLike | None = None,
                      rename_file: str | None = None,
                      cached: bool = False,
                      invalid_cache: bool = False) -> ContextManager[Path]:
    """
    Download file from Google Drive. If not ``cached``, then delete afterward.

    :param file_id: Google Drive file ID used to identify the file to be downloaded.
    :param quiet: Boolean flag to suppress output from the download process.
    :param output_dir: Directory path where the downloaded file will be saved.
    :param rename_file: Optional string to rename the downloaded file.
    :param cached: Boolean flag to retain the downloaded file after usage.
    :param invalid_cache: Boolean flag to force re-download even if the file exists in the cache.
    :return: A context manager yielding the path to the downloaded file.
    """
    if rename_file is not None:
        file = rename_file
    else:
        file = file_id

    if output_dir is None:
        output_dir = NEUROLIB_CACHE_DIRECTORY / 'tmp'

    output_dir.mkdir(exist_ok=True, parents=True)
    output_file = output_dir / file

    try:
        if output_file.exists() and not invalid_cache:
            yield output_file
        else:
            url = f"https://drive.google.com/uc?id={file_id}"
            gdown.download(url, str(output_file), quiet=quiet)
            yield output_file
    finally:
        if not cached:
            output_file.unlink(missing_ok=True)


@contextmanager
def google_drive_folder(folder_id: str,
                        *,
                        quiet: bool = True,
                        output_dir: PathLike | None = None,
                        rename_folder: str | None = None,
                        cached: bool = False,
                        invalid_cache: bool = False) -> ContextManager[Path]:
    """
    Download a entire folder from Google Drive. If not ``cached``, then delete afterward.

    :param folder_id: Google Drive folder ID to download.
    :param quiet: If True, suppress the gdown output.
    :param output_dir: Directory where the folder should be downloaded. If None, a temporary directory is used.
    :param rename_folder: Optional new name for the downloaded folder.
    :param cached: If True, keep the downloaded folder for future use.
    :param invalid_cache: If True, force re-download even if cached data exists.
    :return: A context manager yielding the path to the downloaded folder.
    """
    if rename_folder is not None:
        folder_name = rename_folder
    else:
        folder_name = folder_id

    if output_dir is None:
        output_dir = NEUROLIB_CACHE_DIRECTORY / 'tmp' / folder_name

    try:
        if output_dir.exists() and any(output_dir.iterdir()) and not invalid_cache:
            yield output_dir
        else:
            output_dir.mkdir(exist_ok=True, parents=True)
            gdown.download_folder(id=folder_id, output=str(output_dir), quiet=quiet)
            yield output_dir
    finally:
        if not cached:
            shutil.rmtree(output_dir, ignore_errors=True)


# ======================= #
# 2P Calcium Imaging data #
# ======================= #

def load_example_scanbox(**kwargs) -> SBXInfo:
    """
    :param kwargs: Additional keyword arguments pass to ``google_drive_file`` to customize the loading behavior.
    :return: An instance of ``SBXInfo`` loaded from the specified Google Drive file.
    """
    with google_drive_file('1Gcz_xRVWQJ9QMxq3vzZS8VruSbNiuh_s', **kwargs) as file:
        return SBXInfo.load(file)


def load_example_suite2p(**kwargs) -> Suite2PResult:
    """
    :param kwargs: Additional keyword arguments pass to ``google_drive_folder`` to customize the loading behavior.
    :return: An instance of ``Suite2PResult`` loaded with data from the specified Google Drive folder.
    """
    with google_drive_folder('1iVImr_rIywWhCiBDYhcphcSODaWJrhy7', **kwargs) as suite2p_dir:
        return Suite2PResult.load(suite2p_dir)


# ========== #
# ROIs Atlas #
# ========== #

def load_example_rois(**kwargs) -> pl.DataFrame:
    """
    :param kwargs: Additional keyword arguments pass to ``google_drive_file`` to customize the loading behavior.
    :return: A Polars DataFrame containing the example ROIs data
    """
    with google_drive_file('1cf2r3kcqjENBQMe8YzBQvol8tZgscN4J', **kwargs) as file:
        return pl.read_csv(file)


def load_example_rois_image(**kwargs) -> np.ndarray:
    """
    :param kwargs: Additional arguments to be passed to the `google_drive_file` context manager.
    :return: An example imaging array with labeled ROIs
    """
    with google_drive_file('1-ZFC7Fd6IgwbY6X8oetvpGjZaBBiwg0t', **kwargs) as file:
        return np.array(Image.open(file))


# ========== #
# Ephys data #
# ========== #

def load_ephys_meta(**kwargs):
    with google_drive_file('1gB2MegBAJbobEudx2pZmVe0LBQiU8HFU', **kwargs) as _:
        pass


def load_ephys_data(**kwargs):
    with google_drive_file('1U0xAchQagyXRT72M68fRQ4JsRQeW9q5d', **kwargs) as _:
        pass


def load_npx2_reconstruction(**kwargs) -> pl.DataFrame:
    """
    Example of NeuroPixel2 reconstruction data

    :param kwargs: Additional keyword arguments pass to ``google_drive_file`` to customize the loading behavior.
    :return: A Polars DataFrame containing the example DiI labelled traces ROIs
    """
    with google_drive_file('1fRvMNHhGgh5KP3CgGm6CMFth1qIAmwfh', **kwargs) as file:
        return pl.read_csv(file)


# ========== #
# Cache data #
# ========== #

def load_example_rastermap_2p(**kwargs) -> dict[str, Any]:
    """
    :param kwargs: Additional arguments to be passed to the `google_drive_file` context manager.
    :return: A dictionary containing the 2-photon rastermap data cache
    """
    with google_drive_file('1SuzUhkbcdBY71dCnxjnT4pmQOsiOULBC', **kwargs) as file:
        with file.open('rb') as f:
            return pickle.load(f)


def load_example_rastermap_wfield(**kwargs) -> dict[str, Any]:
    """
    :param kwargs: Additional arguments to be passed to the `google_drive_file` context manager.
    :return: A dictionary containing the wide-field rastermap data cache
    """
    with google_drive_file('1zdZ3ihNPObyA1zVY7knJwVDH8MLXXuYB', **kwargs) as file:
        with file.open('rb') as f:
            return pickle.load(f)


# ========== #
# Behavioral #
# ========== #

def load_example_dlc_h5(**kwargs) -> DeepLabCutResult:
    with google_drive_file('1JNhx6Dpe8beP8vnh0yF3o3vY2DfUM-8A', rename_file='test_dlc.h5', **kwargs) as h5:
        with google_drive_file('1juICYcrXa7Vk-fQSBBSg2QcP9DGyHO2E', rename_file='test_dlc.pickle', **kwargs) as meta:
            return load_dlc_result(h5, meta)


def load_example_dlc_csv(**kwargs) -> DeepLabCutResult:
    with google_drive_file('1R2Ze5xjWlavcKvu45JOH3_QkOD4SSkVN', rename_file='test_dlc.csv', **kwargs) as csv:
        with google_drive_file('1juICYcrXa7Vk-fQSBBSg2QcP9DGyHO2E', rename_file='test_dlc.pickle', **kwargs) as meta:
            return load_dlc_result(csv, meta)



from pathlib import Path

from bokeh.models import ColorBar, LinearColorMapper, GlyphRenderer, TextAreaInput, Div
from bokeh.plotting import figure

__all__ = ['ColorBarView',
           'MsgLog',
           'change_html_format',
           'add_html_format']


class ColorBarView:

    def __init__(self,
                 render: GlyphRenderer,
                 key: str,
                 palette: str,
                 **kwargs):
        """
        Create color map
        :param render: `GlyphRenderer` render
        :param key: key in the `ColumnDataSource`, which indicates the value changes
        :param palette:
        """
        self.render = render
        self.key = key
        self.palette = palette

        self.color_mapper = LinearColorMapper(palette=self.palette, domain=[(self.render, f'{self.key}')])
        self.color_bar = ColorBar(color_mapper=self.color_mapper, width=10, label_standoff=12, **kwargs)

    def insert(self, fig: figure):
        fig.add_layout(self.color_bar, 'right')


def insert_color_bar(fig: figure, render: GlyphRenderer, key: str, palette: str, **kwargs):
    """

    :param fig:
    :param render:
    :param key:
    :param palette:
    :param kwargs:
    :return:
    """
    color_mapper = LinearColorMapper(palette=palette, domain=[(render, key)])
    color_bar = ColorBar(color_mapper=color_mapper, width=10, label_standoff=12, **kwargs)
    fig.add_layout(color_bar)


class MsgLog:
    error_set: bool  # flag if error msg

    def __init__(self, value: str, **kwargs):
        self.error_set = False
        self.message_area = TextAreaInput(title='Log:',
                                          value=value,
                                          rows=5,
                                          cols=100,
                                          disabled=True,
                                          css_classes=['message-area'],
                                          **kwargs)

    def on_message(self, message: str, reset: bool):
        """
        *Example*

        >>> class View(View):
        ...   INSTANCE = None
        ...   def __init__(self, *arg, **kwargs):
        ...         View.INSTANCE = self

        >>> def msg_log(*message: str, reset: bool = False)
        ...     View.INSTANCE.msg_log.on_message(message, reset)

        :param message:
        :param reset: If true, reset the mesasge log `TextAreaInput`
        :return:
        """
        self.message_area.disabled = False
        text = self.message_area.value

        if 'ERR' in message:
            self.error_set = True
        elif reset:
            self.error_set = False

        if self.error_set:
            self.message_area.css_classes = ['message-area-error-msg']
        else:
            self.message_area.css_classes = ['message-area']

        if reset:
            self.message_area.value = str(message)
        else:
            self.message_area.value = text + '\n' + str(message)

        self.message_area.disabled = True

    @property
    def set_div(self) -> Div:
        return Div(text="""
            <style type="text/css">
            div.message-area textarea.bk.bk-input {
                background-color: black !important;
                color: lightgreen !important;
                font-family: monospace !important;
            }
            div.message-area-error-msg textarea.bk.bk-input {
                background-color: black !important;
                color: red !important;
                font-family: monospace !important;
            }                
            div.message-area label.bk {
                       color: lightgreen !important;
                       font-family: monospace !important
                   }
            </style> 
            """)


def change_html_format(input_file: Path, *style: dict[str, str]):
    """change style im embedded html in bokeh with `existing` <style> header"""

    tmp = input_file.with_name('tmp.txt')
    with open(input_file, 'r') as f:
        with tmp.open('w') as out:
            for i, line in enumerate(f):
                for s in style:
                    key = list(s.keys())[0]
                    value = list(s.values())[0]
                    if line.strip().startswith(f'{key}'):
                        out.write(line)
                        out.write(f'\t\t\t{value};\n')
                    else:
                        out.write(line)

    tmp.rename(input_file)


def add_html_format(input_file: Path, style: dict[str, str]):
    """change style im embedded html in bokeh with `new` <style> header"""
    tmp = input_file.with_name('tmp.txt')
    with open(input_file, 'r') as f:
        with tmp.open('w') as out:
            for i, line in enumerate(f):
                if line.strip().startswith('<style>'):
                    key = list(style.keys())[0]
                    value = list(style.values())[0]
                    out.write(line)
                    out.write('\t\t{0} {{\n\t\t\t{1}\n\t\t;}}\n\n'.format(key, value))
                else:
                    out.write(line)

    tmp.rename(input_file)
